from deriva.transfer.download.deriva_download import DerivaDownload, GenericDownloader, DerivaDownloadError, \
    DerivaDownloadConfigurationError, DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
from deriva.transfer.download.deriva_download_cli import DerivaDownloadCLI
from deriva.transfer.upload.deriva_upload import DerivaUpload, GenericUploader, DerivaUploadError, DerivaUploadError, \
    DerivaUploadConfigurationError, DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError
from deriva.transfer.upload.deriva_upload_cli import DerivaUploadCLI



import os
import sys
import json
import traceback
import argparse
from deriva.transfer import GenericDownloader
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
    DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
from deriva.core import BaseCLI, KeyValuePairArgs, format_credential, format_exception, urlparse, __version__


class DerivaDownloadCLI(BaseCLI):
    def __init__(self, description, epilog):

        BaseCLI.__init__(self, description, epilog, __version__, hostname_required=True, config_file_required=True)
        self.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number. Default: 1")
        self.parser.add_argument("path", metavar="<output dir>", help="Path to an output directory.")
        self.parser.add_argument("kwargs", metavar="[key=value key=value ...]",
                                 nargs=argparse.REMAINDER, action=KeyValuePairArgs,
                                 help="Variable length of whitespace-delimited key=value pair arguments used for "
                                      "string interpolation in specific parts of the configuration file. "
                                      "For example: key1=value1 key2=value2")

    @staticmethod
    def download(
               output_path,
               hostname,
               catalog=1,
               token=None,
               envars=None,
               config_file=None,
               credential_file=None):

        assert hostname, "A hostname is required!"
        server = dict()
        server["catalog_id"] = catalog
        if hostname.startswith("http"):
            url = urlparse(hostname)
            server["protocol"] = url.scheme
            server["host"] = url.netloc
        else:
            server["protocol"] = "https"
            server["host"] = hostname

        downloader = GenericDownloader(server,
                                       output_dir=output_path,
                                       envars=envars,
                                       config_file=config_file,
                                       credential_file=credential_file)
        downloader.set_dcctx_cid(__class__.__name__)
        if token:
            downloader.setCredentials(format_credential(token))

        return downloader.download()

    def main(self):
        try:
            args = self.parse_cli()
        except ValueError as e:
            sys.stderr.write(str(e))
            return 2
        if not args.quiet:
            sys.stderr.write("\n")

        try:
            downloaded = DerivaDownloadCLI.download(output_path=os.path.abspath(args.path),
                                                    hostname=args.host,
                                                    catalog=args.catalog,
                                                    token=args.token,
                                                    envars=args.kwargs,
                                                    config_file=args.config_file,
                                                    credential_file=args.credential_file)
            sys.stdout.write("\n%s\n" % (json.dumps(downloaded)))
        except DerivaDownloadAuthenticationError:
            sys.stderr.write(("\n" if not args.quiet else "") +
                             "The requested service requires authentication and a valid login session could not be "
                             "found for the specified host.")
            return 1
        except (DerivaDownloadError, DerivaDownloadConfigurationError, DerivaDownloadAuthorizationError) as e:
            sys.stderr.write(("\n" if not args.quiet else "") + format_exception(e))
            return 1
        except:
            traceback.print_exc()
            return 1
        finally:
            if not args.quiet:
                sys.stderr.write("\n\n")
        return 0



# Tests for the datapath module.
#
# Environment variables:
#  DERIVA_PY_TEST_HOSTNAME: hostname of the test server
#  DERIVA_PY_TEST_CREDENTIAL: user credential, if none, it will attempt to get credentail for given hostname
#  DERIVA_PY_TEST_VERBOSE: set for verbose logging output to stdout

import logging
from operator import itemgetter
import os
import unittest
from deriva.core import DerivaServer, get_credential, ermrest_model as _em
from deriva.core.datapath import DataPathException

try:
    from pandas import DataFrame
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

TEST_EXP_MAX = 100
TEST_EXPTYPE_MAX = 10
TEST_EXP_NAME_FORMAT = "experiment-{}"

hostname = os.getenv("DERIVA_PY_TEST_HOSTNAME")
logger = logging.getLogger(__name__)
if os.getenv("DERIVA_PY_TEST_VERBOSE"):
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


def define_test_schema(catalog):
    """Defines the test schema.

    A 'vocab' schema with an 'experiment_type' term table.
    An 'isa' schema with an 'experiment' table, with 'type' that references the vocab table.
    """
    model = catalog.getCatalogModel()
    vocab = model.create_schema(catalog, _em.Schema.define("Vocab"))
    vocab.create_table(catalog, _em.Table.define_vocabulary("Experiment_Type", "TEST:{RID}"))
    isa = model.create_schema(catalog, _em.Schema.define("ISA"))

    # create 'Experiment' table
    table_def = _em.Table.define(
        "Experiment",
        column_defs=[
            _em.Column.define(cname, ctype) for (cname, ctype) in [
                ('Name', _em.builtin_types.text),
                ('Amount', _em.builtin_types.int4),
                ('Time', _em.builtin_types.timestamptz),
                ('Type', _em.builtin_types.text)
            ]
        ],
        key_defs=[
            _em.Key.define(['Name'])
        ],
        fkey_defs=[
            _em.ForeignKey.define(['Type'], 'Vocab', 'Experiment_Type', ['ID'])
        ]
    )
    isa.create_table(catalog, table_def)

    # create copy of 'Experiment' table
    table_def['table_name'] = 'Experiment_Copy'
    isa.create_table(catalog, table_def)


def _generate_experiment_entities(types, count):
    """Generates experiment entities (content only)

    :param types: type entities to be referenced from entities
    :param count: number of entities to return
    :return: a list of dict objects (experiment entities)
    """
    return [
        {
            "Name": TEST_EXP_NAME_FORMAT.format(i),
            "Amount": i,
            "Time": "2018-01-{}T01:00:00.0".format(1 + (i % 31)),
            "Type": types[i % 10]['ID']
        }
        for i in range(count)
    ]


def populate_test_catalog(catalog):
    """Populate the test catalog."""
    paths = catalog.getPathBuilder()
    logger.debug("Insert experiment types")
    type_table = paths.schemas['Vocab'].tables['Experiment_Type']
    types = type_table.insert([
        {"Name": "{}".format(name), "Description": "NA"} for name in range(TEST_EXPTYPE_MAX)
    ], defaults=['ID', 'URI'])
    logger.debug("Inserting experiments")
    exp = paths.schemas['ISA'].tables['Experiment']
    exp.insert(_generate_experiment_entities(types, TEST_EXP_MAX))


@unittest.skipUnless(hostname, "Test host not specified")
class DatapathTests (unittest.TestCase):
    catalog = None

    @classmethod
    def setUpClass(cls):
        logger.debug("setupUpClass begin")
        credential = os.getenv("DERIVA_PY_TEST_CREDENTIAL") or get_credential(hostname)
        server = DerivaServer('https', hostname, credential)
        cls.catalog = server.create_ermrest_catalog()
        try:
            define_test_schema(cls.catalog)
            populate_test_catalog(cls.catalog)
        except Exception:
            # on failure, delete catalog and re-raise exception
            cls.catalog.delete_ermrest_catalog(really=True)
            raise
        logger.debug("setupUpClass done")

    @classmethod
    def tearDownClass(cls):
        logger.debug("tearDownClass begin")
        cls.catalog.delete_ermrest_catalog(really=True)
        logger.debug("tearDownClass done")

    def setUp(self):
        self.paths = self.catalog.getPathBuilder()
        self.experiment = self.paths.schemas['ISA'].tables['Experiment']
        self.experiment_type = self.paths.schemas['Vocab'].tables['Experiment_Type']
        self.experiment_copy = self.paths.schemas['ISA'].tables['Experiment_Copy']
        self.types = list(self.experiment_type.entities())

    def tearDown(self):
        try:
            self.experiment_copy.path.delete()
        except DataPathException:
            # suppresses 404 errors when the table is empty
            pass

    def test_dir_model(self):
        self.assertIn('ISA', dir(self.paths))

    def test_dir_schema(self):
        self.assertIn('Experiment', dir(self.paths.ISA))

    def test_dir_table(self):
        self.assertIn('Name', dir(self.paths.ISA.Experiment))

    def test_dir_path(self):
        self.assertIn('Experiment', dir(self.paths.ISA.Experiment.path))

    def test_describe_schema(self):
        self.assert_(self.paths.schemas['ISA'].describe())

    def test_describe_table(self):
        self.assert_(self.paths.schemas['ISA'].tables['Experiment'].describe())

    def test_describe_column(self):
        self.assert_(self.paths.schemas['ISA'].tables['Experiment'].column_definitions['Name'].describe())

    def test_unfiltered_fetch(self):
        entities = self.experiment.entities()
        self.assertEquals(len(entities), TEST_EXP_MAX)

    def test_fetch_with_limit(self):
        entities = self.experiment.entities()
        limit = TEST_EXP_MAX / 5
        entities.fetch(limit=limit)
        self.assertEquals(len(entities), limit)

    def test_attribute_projection(self):
        entities = self.experiment.entities(
            self.experiment.column_definitions['Name'],
            self.experiment.column_definitions['Amount']
        )
        entity = entities.fetch(limit=1)[0]
        self.assertIn('Name', entity)
        self.assertIn('Amount', entity)

    def test_link(self):
        entities = self.experiment.link(self.experiment_type).entities()
        self.assertEquals(len(entities), TEST_EXPTYPE_MAX)

    def test_filter_equality(self):
        entities = self.experiment.filter(
            self.experiment.column_definitions['Name'] == TEST_EXP_NAME_FORMAT.format(1)
        ).entities()
        self.assertEquals(len(entities), 1)

    def test_filter_inequality(self):
        entities = self.experiment.filter(
            self.experiment.column_definitions['Amount'] < 10
        ).entities()
        self.assertEquals(len(entities), 10)

    def test_filter_ciregexp(self):
        entities = self.experiment.filter(
            self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
        ).entities()
        self.assertEquals(len(entities), 1)

    def test_filter_negation(self):
        entities = self.experiment.filter(
            ~ (self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:]))
        ).entities()
        self.assertEquals(len(entities), TEST_EXP_MAX - 1)

    def test_filter_conjunction(self):
        entities = self.experiment.filter(
            self.experiment.column_definitions['Name'].ciregexp(TEST_EXP_NAME_FORMAT.format(0)[10:])
            & (self.experiment.column_definitions['Amount'] == 0)
        ).entities()
        self.assertEquals(len(entities), 1)

    def test_attribute_rename(self):
        entities = self.experiment.entities(
            self.experiment.column_definitions['Name'],
            howmuch=self.experiment.column_definitions['Amount']
        )
        entity = entities.fetch(limit=1)[0]
        self.assertIn('Name', entity)
        self.assertIn('howmuch', entity)

    def test_context(self):
        path = self.experiment.link(self.experiment_type)
        entities = path.Experiment.entities()
        self.assertEquals(len(entities), TEST_EXP_MAX)

    def test_path_project(self):
        path = self.experiment.link(self.experiment_type)
        entities = path.Experiment.entities(
            path.Experiment,
            path.Experiment_Type.column_definitions['URI'],
            exptype=path.Experiment_Type.column_definitions['Name']
        )
        entity = entities.fetch(limit=1)[0]
        self.assertIn('Experiment:Name', entity)
        self.assertIn('Experiment:Time', entity)
        self.assertIn('URI', entity)
        self.assertIn('exptype', entity)

    @unittest.skipUnless(HAS_PANDAS, "pandas library not available")
    def test_dataframe(self):
        entities = self.experiment.entities()
        df = entities.dataframe
        self.assertEquals(len(df), TEST_EXP_MAX)

    def test_insert_empty_entities(self):
        entities = self.experiment_copy.insert(None)
        self.assertEqual(len(entities), 0)
        entities = self.experiment_copy.insert([])
        self.assertEqual(len(entities), 0)

    def test_insert_entities_not_iterable(self):
        with self.assertRaises(ValueError):
            self.experiment_type.insert(1)

    def test_insert_entities0_not_dict(self):
        with self.assertRaises(ValueError):
            self.experiment_type.insert([1])
        with self.assertRaises(ValueError):
            self.experiment_type.insert('this is not a dict')

    def test_insert(self):
        entities = self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
        self.assertEqual(len(entities), 10)

    def test_update(self):
        inserted = self.experiment_copy.insert(_generate_experiment_entities(self.types, 10))
        self.assertEqual(len(inserted), 10)
        # now change something in the first entity
        updates = [dict(**inserted[0])]
        updates[0]['Name'] = '**CHANGED**'
        updated = self.experiment_copy.update(updates)
        self.assertEqual(len(updated), 1)
        self.assertEqual(inserted[0]['RID'], updated[0]['RID'])
        self.assertNotEqual(inserted[0]['Name'], updated[0]['Name'])

    def test_update_empty_entities(self):
        entities = self.experiment_copy.update(None)
        self.assertEqual(len(entities), 0)
        entities = self.experiment_copy.update([])
        self.assertEqual(len(entities), 0)

    def test_update_entities_not_iterable(self):
        with self.assertRaises(ValueError):
            self.experiment_type.update(1)

    def test_update_entities0_not_dict(self):
        with self.assertRaises(ValueError):
            self.experiment_type.update([1])
        with self.assertRaises(ValueError):
            self.experiment_type.update('this is not a dict')

    def test_nondefaults(self):
        nondefaults = {'RID', 'RCB', 'RCT'}
        entities = self.experiment.entities()
        self.assertEquals(len(entities), TEST_EXP_MAX)
        entities_copy = self.experiment_copy.insert(entities, nondefaults=nondefaults, add_system_defaults=False)
        self.assertEqual(len(entities), len(entities_copy), 'entities not copied completely')
        ig = itemgetter(*nondefaults)
        for i in range(TEST_EXP_MAX):
            self.assertEqual(ig(entities[i]), ig(entities_copy[i]), 'copied values do not match')

    def test_nondefaults_w_add_sys_defaults(self):
        nondefaults = {'RID', 'RCB', 'RCT'}
        entities = self.experiment.entities()
        self.assertEquals(len(entities), TEST_EXP_MAX)
        entities_copy = self.experiment_copy.insert(entities, nondefaults=nondefaults)
        self.assertEqual(len(entities), len(entities_copy), 'entities not copied completely')
        ig = itemgetter(*nondefaults)
        for i in range(TEST_EXP_MAX):
            self.assertEqual(ig(entities[i]), ig(entities_copy[i]), 'copied values do not match')



import os
import json
import importlib
import logging
from copy import copy
from enum import Enum
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor, \
    LOCAL_PATH_KEY, SOURCE_URL_KEY


class GeoExportTransformProcessor(BaseTransformProcessor):
    def __init__(self, envars=None, **kwargs):
        super(GeoExportTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()

    def process(self):
        logging.info("Transforming input file [%s] into output file [%s]" %
                     (self.input_abspath, self.output_abspath))
        geo = Export2GEO(self.input_abspath, self.output_abspath)
        geo.export_all()

        super(GeoExportTransformProcessor, self).process()
        return self.outputs


class Style(Enum):
    DEFAULT = 0
    INSTRUCTION = 1
    SECTION = 2
    FIELD = 3
    HEADER = 4
    PREFACE = 5


class Export2Excel(object):
    OPENPYXL = None

    def __init__(self, output_path):
        self.__name__ = 'export_excel'
        self.wb = None
        self.ws = None
        self._output_path_ = output_path
        self.sheet_name = 'METADATA'
        self.initialize_wb()

    def import_openpyxl(self):
        # locate library
        if self.OPENPYXL is None:
            try:
                self.OPENPYXL = importlib.import_module("openpyxl")
            except ImportError as e:
                raise DerivaDownloadConfigurationError("Unable to find required module. "
                                                       "Ensure that the Python package \"openpyxl\" is installed.", e)

    def initialize_wb(self):
        self.import_openpyxl()
        # Open an xlsx
        self.wb = self.OPENPYXL.Workbook()
        self.ws = self.wb.create_sheet(index=0, title=self.sheet_name)
        #remove the default sheet
        self.wb.remove(self.wb['Sheet'])

        fgColor0 = self.OPENPYXL.styles.colors.Color(indexed=10, type='indexed')
        fgColor1 = self.OPENPYXL.styles.colors.Color(indexed=42, type='indexed')
        bgColor1 = self.OPENPYXL.styles.colors.Color(indexed=27, type='indexed')
        fgColor2 = self.OPENPYXL.styles.colors.Color(indexed=12, type='indexed')
        fgColor3 = self.OPENPYXL.styles.colors.Color(indexed=13, type='indexed')
        bgColor3 = self.OPENPYXL.styles.colors.Color(indexed=34, type='indexed')

        self.font_red = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor0, sz=10.0)
        self.font_blue = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor2, sz=10.0)
        self.font_bond = self.OPENPYXL.styles.Font(name='Arial', b=True, sz=10.0)

        self.fill_green = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor1, bgColor=bgColor1)
        self.fill_yellow = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor3, bgColor=bgColor3)

        self.ws.column_dimensions['A'].width = 42.0
        self.ws.column_dimensions['B'].width = 15.0
        self.ws.column_dimensions['C'].width = 24.0
        self.ws.column_dimensions['D'].width = 15.0
        self.ws.column_dimensions['E'].width = 24.0
        self.ws.column_dimensions['F'].width = 24.0
        self.ws.column_dimensions['G'].width = 24.0
        self.ws.column_dimensions['H'].width = 24.0
        self.ws.column_dimensions['I'].width = 24.0
        self.ws.column_dimensions['J'].width = 24.0

    @staticmethod
    def cell_name(row_num, col_num):
        string = ""
        while col_num > 0:
            col_num, remainder = divmod(col_num - 1, 26)
            string = chr(65 + remainder) + string
        return string + str(row_num)

    def insert_row(self, row_idx, row_amt):
        self.ws.insert_rows(idx=row_idx, amount=row_amt)

    def delete_row(self, idx, row_amt):
        self.ws.delete_rows(idx=idx, amount=row_amt)

    def write_cell(self, row_idx, col_idx, content, export_style=Style.DEFAULT):
        cell_loc = self.cell_name(row_idx, col_idx)
        cell = self.ws[cell_loc]
        cell.value = content
        if(export_style == Style.INSTRUCTION):
            cell.fill = self.fill_green
        elif(export_style == Style.HEADER):
            cell.fill = self.fill_yellow
            cell.font = self.font_blue
        elif(export_style == Style.FIELD):
            cell.font = self.font_blue
        elif(export_style == Style.SECTION):
            cell.font = self.font_red
        elif(export_style == Style.PREFACE):
            cell.fill = self.fill_green
            cell.font = self.font_bond
        else:
            pass

    def save_xlsx(self):
        self.wb.save(self._output_path_)
    def copy_cell(self, old_row_idx, old_col_idx, new_row_idx, new_col_idx,copy_style=True):
        cell = self.ws[self.cell_name(old_row_idx, old_col_idx)]
        new_cell = self.ws[self.cell_name(new_row_idx, new_col_idx)]
        new_cell.value = cell.value
        if cell.has_style and copy_style:
            new_cell.font = copy(cell.font)
            new_cell.border = copy(cell.border)
            new_cell.fill = copy(cell.fill)
            new_cell.number_format = copy(cell.number_format)
            new_cell.protection = copy(cell.protection)
            new_cell.alignment = copy(cell.alignment)
            new_cell.comment = copy(cell.comment)


class Export2GEO(object):

    def __init__(self, input_file, output_excel_path):
        self.__name__ = 'export_geo_template'
        self.input_file = input_file
        self.output_excel_path = output_excel_path
        self.current_row_idx = 0
        self.header_row_idx = 0
        self.protocol_type = ['Isolation_Protocol', 'Growth_Protocol', 'Treatment_Protocol', 'Extract_Protocol',
                              'Construction_Protocol', 'Label_Protocol', 'Hybridization_Protocol', 'Scan_Protocol']
        self.protocol_unique = {}
        self.protocol_list = {}
        self.other_item = ['Data_Processing', 'Reference_Genome']
        self.other_item_unique = {}
        self.other_item_list = {}
        self.initialize_data()

    def initialize_data(self):
        with open(self.input_file) as input_file:
            data = json.load(input_file)
            self.study = None
            if data is None or len(data) == 0:
                logging.debug("No Data")
            elif len(data) > 1:
                logging.debug("Duplicate Data")
            elif len(data) == 1:
                self.study = data[0].get('Study')[0]
                self.pi = data[0].get('Principal_Investigator')[0]
                self.experiments = data[0].get('Experiment')
                self.experiment_settings = data[0].get('Experiment_Setting')
                self.replicates = data[0].get('Replicate')
                self.specimen = data[0].get('Specimen')
                self.files = data[0].get('File')
                self.study_files = data[0].get('Study_File')
                self.stage = data[0].get('Stage')
                self.tissue = data[0].get('Tissue')
                self.source_type = data[0].get('Source')
                self.specimen_cell_type = data[0].get('Specimen_Cell_Type')
                self.cell_type = data[0].get('Cell_Type')

    def export_start(self):
        self.excel = Export2Excel(self.output_excel_path)

    def export_all(self):
        if self.study is not None:
            self.export_start()
            self.export_preface()
            self.export_series()
            self.export_sample_prepare()
            self.export_sample()
            self.export_protocol()
            self.export_data_processing()
            self.export_processed_datafiles()
            self.export_raw_datafiles()
            self.export_paired_end()
            self.export_finish()

    def export_preface(self):
        self.current_row_idx = 1
        self.excel.write_cell(self.current_row_idx, 1, '# High-throughput sequencing metadata template (version 2.1).',Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# All fields in this template must be completed.',Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Templates containing example data are found in the METADATA EXAMPLES spreadsheet tabs at the foot of this page.',Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Field names (in blue on this page) should not be edited. Hover over cells containing field names to view field content guidelines.',Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# Human data. If there are patient privacy concerns regarding making data fully public through GEO, '
        'please submit to NCBI''s dbGaP (http://www.ncbi.nlm.nih.gov/gap/) database. dbGaP has controlled access'
        'mechanisms and is an appropriate resource for hosting sensitive patient data.'),Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '',Style.PREFACE)
        self.current_row_idx += 1

        for col in range(2,13):
            for row in range(1,7):
                self.excel.write_cell(row, col, '', Style.PREFACE)
    # write SERIES

    def export_series(self):
        self.excel.write_cell(self.current_row_idx, 1, 'SERIES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# This section describes the overall experiment.', Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'title', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Title',''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'summary', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Summary',''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'overall design', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Overall_Design',''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'contributor', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.pi.get('Full_Name',''))
        self.current_row_idx += 1

        if len(self.study_files) >= 1 and self.study_files[0] is not None:
            self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, self.study_files[0].get('File_Name',''))
            self.current_row_idx += 1
            for i in range(len(self.study_files) - 1):
                self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, self.study_files[i + 1].get('File_Name',''))
                self.current_row_idx += 1
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
            logging.debug('No study file')

    def export_sample_prepare(self):
        # check self.protocol unique
        for p in self.protocol_type:
            # protocols which are applicable to only a subset of Samples should be included
            # as additional columns of the SAMPLES section instead.
            # this dictionary is used to handle dynamic headers related to protocol
            self.protocol_unique[p] = True
            self.protocol_list[p] = []

        for i in self.other_item:
            self.other_item_unique[i] = True
            self.other_item_list[i] = []

        for e in self.experiments:
            if e is None or 'RID' not in e.keys():
                continue
            for p in self.protocol_type:
                if p not in e.keys():
                    continue
                elif len(self.protocol_list[p]) == 0 and e[p] is not None and e[p] != 'None':
                    self.protocol_list[p].append(e[p])
                elif e[p] not in self.protocol_list[p] and e[p] is not None and e[p] != 'None':
                    self.protocol_list[p].append(e[p])
                    self.protocol_unique[p] = False
            i = 'Data_Processing'
            if i in e.keys():
                if len(self.other_item_list[i]) == 0 and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                elif e[i] not in self.other_item_list[i] and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                    self.other_item_unique[i] = False

        for e in self.experiment_settings:
            if e is None:
                continue
            i = 'Reference_Genome'
            if i in e.keys():
                if len(self.other_item_list[i]) == 0 and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                elif e[i] not in self.other_item_list[i] and e[i] is not None and e[i] != 'None':
                    self.other_item_list[i].append(e[i])
                    self.other_item_unique[i] = False

    def export_sample(self):
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'SAMPLES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# This section lists and describes each of the biological Samples under investgation,'
        'as well as any protocols that are specific to individual Samples.'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Additional "processed data file" or "raw file" columns may be included.', Style.INSTRUCTION)

        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1

        # write SAMPLES
        for e in self.experiments:
            if e is None or 'RID' not in e.keys():
                continue
            for r in self.replicates:
                if r is None or 'Experiment_RID' not in r.keys() or 'RID' not in r.keys():
                    continue
                elif r['Experiment_RID'] == e['RID']:
                    local_col_idx = 1
                    self.current_row_idx += 1
                    sample_name = e.get('Internal_ID','') + '_' + str(r.get('Biological_Replicate_Number','')) + '_' + str(
                        r.get('Technical_Replicate_Number',''))
                    sample_title = e.get('Name','') + '_' + str(r.get('Biological_Replicate_Number','')) + '_' + str(
                        r.get('Technical_Replicate_Number',''))
                    sample_source_name = e.get('Anatomy','')
                    sample_organism = e.get('Species','')
                    sample_molecule = e.get('Molecule_Type','')
                    sample_description = r.get('Notes', '')

                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'Sample name', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_name)
                    local_col_idx += 1
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'title', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_title)
                    local_col_idx += 1
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'source name', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_source_name)
                    local_col_idx += 1
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'organism', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_organism)
                    local_col_idx += 1

                    for p in self.protocol_type:
                        if not self.protocol_unique[p]:
                            self.excel.write_cell(self.header_row_idx, local_col_idx, (p.replace('_', ' ').lower()),Style.HEADER)
                            self.excel.write_cell(self.current_row_idx, local_col_idx, e[p])
                            local_col_idx += 1

                    characteristic_list = ['Phenotype','Stage_ID','Stage_Detail','Genotype','Strain','Wild_Type']
                    foreign_item_list = ['Tissue','Source','Specimen_Cell_Type','Cell_Type']
                    characteristic_exist=[]
                    for s in self.specimen:
                        for c in characteristic_list:
                            characteristic = s.get(c,'')
                            if characteristic is not None and len(characteristic)>0 and c not in characteristic_exist:
                                characteristic_exist.append(c)
                            else:
                                continue

                    for f in self.source_type:
                        if f is None:
                            continue
                        else:
                            characteristic_exist.append('Source')
                            break
                    for f in self.cell_type:
                        if f is None:
                            continue
                        else:
                            characteristic_exist.append('Cell_Type')
                            break

                    if 'Stage_ID' in characteristic_exist and 'Stage_Detail' in characteristic_exist:
                        characteristic_exist.remove('Stage_ID')
                        characteristic_exist.remove('Stage_Detail')
                        characteristic_exist.append('Stage')
                    elif 'Stage_ID' in characteristic_exist:
                        characteristic_exist.remove('Stage_ID')
                        characteristic_exist.append('Stage')
                    elif 'Stage_Detail' in characteristic_exist:
                        characteristic_exist.remove('Stage_Detail')
                        characteristic_exist.append('Stage')

                    for s in self.specimen:
                        if s is None or 'REPLICATE_RID' not in s.keys():
                            continue
                        elif s['REPLICATE_RID'] == r['RID']:
                            for c in characteristic_exist:
                                #handling characteristic source_type
                                if c == "Source":
                                    source_type = ''
                                    for p in self.tissue:
                                        for q in self.source_type:
                                            if p is None or q is None:
                                                continue
                                            elif s['RID'] == p['Specimen_RID'] and p['Tissue'] == q['ID']:
                                                source_type = source_type + q['Name']+','
                                            else:
                                                continue
                                    characteristic = source_type[:-1] if len(source_type)>0 else ''
                                    self.excel.write_cell(self.header_row_idx, local_col_idx,
                                                          'characteristics: Source',
                                                          Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1
                                # handling characteristic cell_type
                                elif c == "Cell_Type":
                                    cell_type = ''
                                    for p in self.specimen_cell_type:
                                        for q in self.cell_type:
                                            if p is None or q is None:
                                                continue
                                            elif s['RID'] == p['Specimen'] and p['Cell_Type']==q['Name']:
                                                cell_type = cell_type + q['Name'] + ','
                                            else:
                                                continue
                                    characteristic = cell_type[:-1] if len(cell_type)>0 else ''
                                    self.excel.write_cell(self.header_row_idx, local_col_idx,
                                                          'characteristics: Cell_Type',
                                                          Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1
                                # handling characteristic stage
                                elif c == "Stage":
                                    get_stage = 0
                                    for v in self.stage:
                                        if v is None:
                                            continue
                                        elif v['ID'] == s['Stage_ID']:
                                            characteristic = v['Name']
                                            self.excel.write_cell(self.header_row_idx, local_col_idx,
                                                                  'characteristics: Stage',
                                                                  Style.HEADER)
                                            self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                            local_col_idx += 1
                                            get_stage = 1
                                        else:
                                            continue
                                    # using Stage_Detail if no Stage_ID
                                    if get_stage == 0:
                                        characteristic = s.get('Stage_Detail', '')
                                        self.excel.write_cell(self.header_row_idx, local_col_idx,
                                                              'characteristics: ' + 'Stage',
                                                              Style.HEADER)
                                        self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                        local_col_idx += 1
                                else:
                                    characteristic = s.get(c, '')
                                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'characteristics: '+c,
                                                          Style.HEADER)
                                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                                    local_col_idx += 1
                        else:
                            continue

                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'molecule',Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_molecule)
                    local_col_idx += 1

                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'description', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, sample_description)
                    local_col_idx += 1

                    # processed files are write to supplementary file

                    # writing raw file
                    for f in self.files:
                        if f is None or 'Replicate_RID' not in f.keys():
                            continue
                        elif f['Replicate_RID'] == r['RID']:
                            self.excel.write_cell(self.header_row_idx, local_col_idx, 'raw file', Style.HEADER)
                            self.excel.write_cell(self.current_row_idx, local_col_idx, f.get('File_Name',''))
                            local_col_idx += 1
                        else:
                            continue
                else:
                    continue

    def export_protocol(self):
        self.current_row_idx = self.current_row_idx + 2
        self.excel.write_cell(self.current_row_idx, 1, 'PROTOCOLS', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# Any of the protocols below which are applicable to only a subset of '
        'Samples should be included as additional columns of the SAMPLES section instead.'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'growth protocol', Style.FIELD)
        if self.protocol_unique['Growth_Protocol']:
            if len(self.protocol_list['Growth_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Growth_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'treatment protocol', Style.FIELD)
        if self.protocol_unique['Treatment_Protocol']:
            if len(self.protocol_list['Treatment_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Treatment_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'extract protocol', Style.FIELD)
        if self.protocol_unique['Extract_Protocol']:
            if len(self.protocol_list['Extract_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Extract_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'library construction protocol', Style.FIELD)
        if self.protocol_unique['Construction_Protocol']:
            if len(self.protocol_list['Construction_Protocol']) == 1:
                self.excel.write_cell(self.current_row_idx, 2, self.protocol_list['Construction_Protocol'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 2, '')
        else:
            self.excel.write_cell(self.current_row_idx, 2, '')
        self.current_row_idx += 1

        for p in self.protocol_type:
            if p not in ['Growth_Protocol', 'Treatment_Protocol', 'Extract_Protocol', 'Construction_Protocol']:
                if self.protocol_unique[p] and len(self.protocol_list[p]) == 1:
                    self.excel.write_cell(self.current_row_idx, 1, (p.replace('_', ' ').lower()),Style.FIELD)
                    self.excel.write_cell(self.current_row_idx, 2, self.protocol_list[p][0])
                    self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'library strategy',Style.FIELD)
        # todo: fixed value for now.
        self.excel.write_cell(self.current_row_idx, 2, 'RNA-seq')

    def export_data_processing(self):
        # DATA PROCESSING PIPELINE
        self.current_row_idx = self.current_row_idx + 2

        self.excel.write_cell(self.current_row_idx, 1, 'DATA PROCESSING PIPELINE', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# Data processing steps include base-calling, alignment, filtering,'
'peak-calling, generation of normalized abundance measurements etc'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For each step provide a description, as well as software name,'
        'version, parameters, if applicable.'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Include additional steps, as necessary.', Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1

        if len(self.other_item_list['Data_Processing']) == 1:
            processing_list = self.other_item_list['Data_Processing'][0].split('\n')
        else:
            processing_list = []

        other_procession_set = {}
        for es in self.experiment_settings:
            if es is None or 'Experiment_RID' not in es.keys():
                continue
            else:
                if other_procession_set:
                    for k,v in es.items():
                        if k in other_procession_set.keys() and v != other_procession_set[k]:
                            del other_procession_set[k]
                else:
                    other_procession_set = es.copy()


        for step_row in processing_list:
            self.excel.write_cell(self.current_row_idx, 1, 'data processing step',Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, step_row)
            self.current_row_idx += 1

        for k, v in other_procession_set.items():
            if v and k not in ['RCB','RCT','RID','RMB','RMT'] and len(v)>0:
                self.excel.write_cell(self.current_row_idx, 1, 'data processing step', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, k+' : '+v)
                self.current_row_idx += 1

        if self.other_item_unique['Reference_Genome']:
            if len(self.other_item_list['Reference_Genome']) == 1:
                self.excel.write_cell(self.current_row_idx, 1, 'genome build', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, self.other_item_list['Reference_Genome'][0])
            else:
                self.excel.write_cell(self.current_row_idx, 1, 'genome build', Style.FIELD)
                self.excel.write_cell(self.current_row_idx, 2, ' '.join(self.other_item_list['Reference_Genome']))
                logging.debug('Multi-genome')

        self.current_row_idx += 1

        self.excel.write_cell(self.current_row_idx, 1, 'processed data files format and content',Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, 'processed data file format')
        self.current_row_idx += 1

    def export_processed_datafiles(self):
        # PROCESSED DATA FILES
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For each file listed in the "processed data file" columns of '
        'the SAMPLES section, provide additional information below.'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PROCESSED DATA FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2

        for pf in self.study_files:
            if pf is None:
                continue
            else:
                local_col_idx = 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('File_Name',''))
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('File_Type',''))
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('MD5',''))
                self.current_row_idx += 1

    def export_raw_datafiles(self):
        # RAW FILES
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For each file listed in the "raw file" columns of'
        'the SAMPLES section, provide additional information below.'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'RAW FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2

        for pf in self.files:
            if pf is None or 'Experiment_RID' not in pf.keys():
                continue
            else:
                current_col_idx = 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('File_Name',''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('File_Type',''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('MD5',''))
                current_col_idx += 1
                # Experiment_Settings.Protocol_Reference
                for es in self.experiment_settings:
                    if es is None or 'Experiment_RID' not in es.keys():
                        print('skipped?')
                        print(es)
                        continue
                    elif es['Experiment_RID'] == pf['Experiment_RID']:
                        self.excel.write_cell(self.header_row_idx, current_col_idx, 'instrument model', Style.HEADER)
                        self.excel.write_cell(self.current_row_idx, current_col_idx, es.get('Sequencing_Platform',''))
                        current_col_idx += 1
                        self.excel.write_cell(self.header_row_idx, current_col_idx, 'read length', Style.HEADER)
                        self.excel.write_cell(self.current_row_idx, current_col_idx, es.get('Read_Length',''))
                        current_col_idx += 1
                        if 'pair' in str(es.get('Paired_End','')).lower():
                            single_or_paired = 'paired-end'
                        else:
                            single_or_paired = 'single'
                        self.excel.write_cell(self.header_row_idx, current_col_idx, 'single or paired-end', Style.HEADER)
                        self.excel.write_cell(self.current_row_idx, current_col_idx, single_or_paired)
                        current_col_idx += 1
            self.current_row_idx += 1

    def export_paired_end(self):
        # PAIRED-END EXPERIMENTS
        # todo need check what's paired-end data look like
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For paired-end experiments, list the 2 associated raw files, and provide average insert size and standard deviation,'
        'if known. For SOLiD experiments, list the 4 file names (include "file name 3" and "file name 4" columns).'), Style.INSTRUCTION)
        for col in range(2,13):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PAIRED-END EXPERIMENTS', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1
        current_col_idx = 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 1', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 2', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'average insert size', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'standard deviation', Style.HEADER)

    def export_finish(self):
        self.excel.save_xlsx()


import io
import os
import sys
import shutil
import errno
import json
import platform
import logging
import requests
import inspect
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.packages.urllib3.exceptions import MaxRetryError
from collections import OrderedDict
from distutils import util as du_util
from importlib import import_module

__version__ = "0.8.1"

IS_PY2 = (sys.version_info[0] == 2)
IS_PY3 = (sys.version_info[0] == 3)

if IS_PY3:
    from urllib.parse import quote as _urlquote, unquote as urlunquote
    from urllib.parse import urlparse, urlsplit, urlunsplit
    from http.cookiejar import MozillaCookieJar
else:
    from urllib import quote as _urlquote, unquote as urlunquote
    from urlparse import urlparse, urlsplit, urlunsplit
    from cookielib import MozillaCookieJar


def urlquote(s, safe=''):
    """Quote all reserved characters according to RFC3986 unless told otherwise.

       The urllib.urlquote has a weird default which excludes '/' from
       quoting even though it is a reserved character.  We would never
       want this when encoding elements in Deriva REST API URLs, so
       this wrapper changes the default to have no declared safe
       characters.

    """
    return _urlquote(s.encode('utf-8'), safe=safe)


def urlquote_dcctx(s, safe='~{}",:'):
    """Quote for use with Deriva-Client-Context or other HTTP headers.

       Defaults to allow additional safe characters for less
       aggressive encoding of JSON content for use in an HTTP header
       value.

    """
    return urlquote(s, safe=safe)


DEFAULT_HEADERS = {}

Kilobyte = 1024
Megabyte = Kilobyte ** 2
DEFAULT_CHUNK_SIZE = Megabyte * 10  # above the minimum 5MB chunk size for AWS S3 multipart uploads
MAX_CHUNK_SIZE = Megabyte * 100


class NotModified (ValueError):
    pass


class ConcurrentUpdate (ValueError):
    pass


def stob(string):
    return bool(du_util.strtobool(str(string)))


def format_exception(e):
    exc = "".join(("[", type(e).__name__, "] "))
    if isinstance(e, requests.HTTPError):
        resp = " - Server responded: %s" % e.response.text.strip().replace('\n', ': ')
        return "".join((exc, str(e), resp))
    return "".join((exc, str(e)))


def frozendict(d):
    items = list(d.items())
    for i in range(len(items)):
        k, v = items[i]
        if isinstance(v, dict):
            v = frozendict(v)
            items[i] = (k, v)
    return frozenset(set(items))


def add_logging_level(level_name, level_num, method_name=None):
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        logging.warning('{} already defined in logging module'.format(level_name))
        return
    if hasattr(logging, method_name):
        logging.warning('{} already defined in logging module'.format(method_name))
        return
    if hasattr(logging.getLoggerClass(), method_name):
        logging.warning('{} already defined in logger class'.format(method_name))
        return

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def init_logging(level=logging.INFO,
                 log_format="%(asctime)s - %(levelname)s - %(message)s",
                 file_path=None,
                 capture_warnings=True):
    add_logging_level("TRACE", logging.DEBUG-5)
    logging.captureWarnings(capture_warnings)
    # this will suppress potentially numerous INFO-level "Resetting dropped connection" messages from requests
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)
    if file_path:
        logging.basicConfig(filename=file_path, level=level, format=log_format)
    else:
        logging.basicConfig(level=level, format=log_format)


DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.deriva')
DEFAULT_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'credential.json')
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'config.json')
DEFAULT_COOKIE_JAR_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'cookies.txt')
DEFAULT_SESSION_CONFIG = {
    "retry_connect": 5,
    "retry_read": 5,
    "retry_backoff_factor": 1.0,
    "retry_status_forcelist": [500, 503, 504],
    "cookie_jar": DEFAULT_COOKIE_JAR_FILE
}
DEFAULT_CONFIG = {
    "server":
    {
        "protocol": "https",
        "host": platform.uname()[1],
        "catalog_id": 1
    },
    "session": DEFAULT_SESSION_CONFIG,
    "download_processor_whitelist": []
}

DEFAULT_CREDENTIAL = {}


def get_new_requests_session(url=None, session_config=DEFAULT_SESSION_CONFIG):
    session = requests.session()
    inspect_retry = inspect.getargspec(Retry.__init__)
    if "raise_on_status" in inspect_retry.args:
        retries = Retry(connect=session_config['retry_connect'],
                        read=session_config['retry_read'],
                        backoff_factor=session_config['retry_backoff_factor'],
                        status_forcelist=session_config['retry_status_forcelist'],
                        method_whitelist=False,
                        raise_on_status=True)
    else:
        # this is in case installed urllib3 is < 1.15 and raise_on_status is unavailable
        retries = Retry(connect=session_config['retry_connect'],
                        read=session_config['retry_read'],
                        backoff_factor=session_config['retry_backoff_factor'],
                        status_forcelist=session_config['retry_status_forcelist'])
    if url:
        session.mount(url, HTTPAdapter(max_retries=retries))
    else:
        session.mount('http://', HTTPAdapter(max_retries=retries))
        session.mount('https://', HTTPAdapter(max_retries=retries))

    return session


def make_dirs(path, mode=0o777):
    if not os.path.isdir(path):
        try:
            os.makedirs(path, mode=mode)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise


def copy_config(src, dst):
    config_dir = os.path.dirname(dst)
    make_dirs(config_dir, mode=0o750)
    shutil.copy2(src, dst)


def write_config(config_file=DEFAULT_CONFIG_FILE, config=DEFAULT_CONFIG):
    config_dir = os.path.dirname(config_file)
    make_dirs(config_dir, mode=0o750)
    with io.open(config_file, 'w', newline='\n', encoding='utf-8') as cf:
        config_data = json.dumps(config, ensure_ascii=False, indent=2)
        if IS_PY2 and isinstance(config_data, str):
            config_data = unicode(config_data, 'utf-8')
        cf.write(config_data)
        cf.close()


def read_config(config_file=DEFAULT_CONFIG_FILE, create_default=False, default=DEFAULT_CONFIG):
    if not config_file:
        config_file = DEFAULT_CONFIG_FILE
    config = None
    if not os.path.isfile(config_file) and create_default:
        logging.info("No default configuration file found, attempting to create one at: %s" % config_file)
        try:
            write_config(config_file, default)
        except Exception as e:
            logging.warning("Unable to create configuration file %s. Using internal defaults. %s" %
                            (config_file, format_exception(e)))
            config = default

    if not config:
        with open(config_file) as cf:
            config = cf.read()

    return json.loads(config, object_pairs_hook=OrderedDict)


PORTALOCKER = None
try:
    PORTALOCKER = import_module("portalocker")
except ImportError:
    pass


def lock_file(file_path, mode, exclusive=True):
    global PORTALOCKER
    if PORTALOCKER and os.path.isfile(file_path):
        portalocker_argspec = inspect.getargspec(PORTALOCKER.Lock.__init__)
        if "truncate" not in portalocker_argspec.args:
            return PORTALOCKER.Lock(file_path, mode, timeout=60,
                                    flags=PORTALOCKER.LOCK_EX if exclusive else PORTALOCKER.LOCK_SH)
        else:
            return PORTALOCKER.Lock(file_path, mode, timeout=60,
                                    flags=PORTALOCKER.LOCK_EX if exclusive else PORTALOCKER.LOCK_SH, truncate=None)
    else:
        return io.open(file_path, mode)


def write_credential(credential_file=DEFAULT_CREDENTIAL_FILE, credential=DEFAULT_CREDENTIAL):
    credential_dir = os.path.dirname(credential_file)
    make_dirs(credential_dir, mode=0o750)
    with lock_file(credential_file, mode='w', exclusive=True) as cf:
        os.chmod(credential_file, 0o600)
        credential_data = json.dumps(credential, ensure_ascii=False, indent=2)
        if IS_PY2 and isinstance(credential_data, str):
            credential_data = unicode(credential_data, 'utf-8')
        cf.write(credential_data)
        cf.flush()
        cf.close()


def read_credential(credential_file=DEFAULT_CREDENTIAL_FILE, create_default=False, default=DEFAULT_CREDENTIAL):
    if not credential_file:
        credential_file = DEFAULT_CREDENTIAL_FILE
    credential = None
    if not os.path.isfile(credential_file) and create_default:
        logging.info("No default credential file found, attempting to create one at: %s" % credential_file)
        try:
            write_credential(credential_file, default)
        except Exception as e:
            logging.warning("Unable to create configuration file %s. Using internal defaults. %s" %
                            (credential_file, format_exception(e)))
            credential = default

    if not credential:
        with lock_file(credential_file, mode='r') as cf:
            credential = cf.read()

    return json.loads(credential, object_pairs_hook=OrderedDict)


def get_credential(host, credential_file=DEFAULT_CREDENTIAL_FILE):
    if credential_file is None:
        credential_file = DEFAULT_CREDENTIAL_FILE
    credentials = read_credential(credential_file)
    return credentials.get(host, credentials.get(host.lower()))


def format_credential(token=None, username=None, password=None):
    if token:
        return {"cookie": "webauthn=%s" % token}
    elif username and password:
        return {"username": username, "password": password}
    raise ValueError(
        "Missing required argument(s): an authentication token or a username and password must be provided.")


def bootstrap():
    init_logging()
    read_config(create_default=True)
    read_credential(create_default=True)


def load_cookies_from_file(cookie_file=None):
    if not cookie_file:
        cookie_file = DEFAULT_SESSION_CONFIG["cookie_jar"]
    cookies = MozillaCookieJar()
    if os.path.isfile(cookie_file):
        # Load saved cookies
        try:
            cookies.load(cookie_file, ignore_discard=True, ignore_expires=True)
            return cookies
        except Exception as e:
            logging.warning(format_exception(e))
    # Create cookie file
    cookies.save(cookie_file, ignore_discard=True, ignore_expires=True)
    os.chmod(cookie_file, 0o600)

    return cookies


def resource_path(relative_path, default=os.path.abspath(".")):
    # required to find bundled data at runtime in Pyinstaller single-file exe mode
    if getattr(sys, 'frozen', False):
        return os.path.join(getattr(sys, '_MEIPASS', '.'), relative_path)
    if default is None:
        return relative_path
    return os.path.join(default, relative_path)


def get_transfer_summary(total_bytes, elapsed_time):
    total_secs = elapsed_time.total_seconds()
    transferred = \
        float(total_bytes) / float(Kilobyte) if total_bytes < Megabyte else float(total_bytes) / float(Megabyte)
    throughput = str(" at %.2f MB/second" % (transferred / total_secs)) if (total_secs >= 1) else ""
    elapsed = str("Elapsed time: %s." % elapsed_time) if (total_secs > 0) else ""
    summary = "%.2f %s transferred%s. %s" % \
              (transferred, "KB" if total_bytes < Megabyte else "MB", throughput, elapsed)
    return summary


from deriva.core.base_cli import BaseCLI, KeyValuePairArgs
from deriva.core.deriva_binding import DerivaBinding, DerivaPathError, DerivaClientContext
from deriva.core.deriva_server import DerivaServer
from deriva.core.ermrest_catalog import ErmrestCatalog, ErmrestSnapshot, ErmrestCatalogMutationError
from deriva.core.ermrest_config import AttrDict, CatalogConfig
from deriva.core.polling_ermrest_catalog import PollingErmrestCatalog
from deriva.core.hatrac_store import HatracStore, HatracHashMismatch, HatracJobPaused, HatracJobAborted, \
    HatracJobTimeout




from . import ermrest_config as _ec
import re

def _kwargs(**kwargs):
    """Helper for extending ermrest_config with sub-types for the whole model tree."""
    kwargs2 = {
        'schema_class': Schema,
        'table_class': Table,
        'column_class': Column,
        'key_class': Key,
        'foreign_key_class': ForeignKey,
    }
    kwargs2.update(kwargs)
    return kwargs2

class Model (_ec.CatalogConfig):
    """Top-level catalog model.
    """
    def __init__(self, model_doc, **kwargs):
        super(Model, self).__init__(model_doc, **_kwargs(**kwargs))
        self.update_referenced_by()

    def update_referenced_by(self):
        """Introspects the 'foreign_keys' and updates the 'referenced_by' properties on the 'Table' objects.
        :param model: an ERMrest model object
        """
        for schema in self.schemas.values():
            for referer in schema.tables.values():
                for fkey in referer.foreign_keys:
                    referenced = self.schemas[
                        fkey.referenced_columns[0]['schema_name']
                    ].tables[
                        fkey.referenced_columns[0]['table_name']
                    ]
                    referenced.referenced_by.append(fkey)

    def create_schema(self, catalog, schema_def):
        """Add a new schema to this model in the remote database based on schema_def.

           Returns a new Schema instance based on the server-supplied
           representation of the newly created schema.

           The returned Schema is also added to self.schemas.
        """
        sname = schema_def['schema_name']
        if sname in self.schemas:
            raise ValueError('Schema %s already exists.' % sname)
        r = catalog.post(
            self.uri_path,
            json=[schema_def],
        )
        r.raise_for_status()
        d = r.json()
        assert len(d) == 1
        newschema = Schema(sname, d[0])
        self.schemas[sname] = newschema
        return newschema

class Schema (_ec.CatalogSchema):
    """Named schema.
    """
    def __init__(self, sname, schema_doc, **kwargs):
        super(Schema, self).__init__(sname, schema_doc, **_kwargs(**kwargs))
        self.comment = schema_doc.get('comment')

    @classmethod
    def define(cls, sname, comment=None, acls={}, annotations={}):
        """Build a schema definition.
        """
        return {
            "schema_name": sname,
            "acls": acls,
            "annotations": annotations,
            "comment": comment,
        }

    def prejson(self, prune=True):
        d = super(Schema, self).prejson(prune)
        d.update({
            'comment': self.comment,
        })
        return d

    def create_table(self, catalog, table_def):
        """Add a new table to this schema in the remote database based on table_def.

           Returns a new Table instance based on the server-supplied
           representation of the newly created table.

           The returned Table is also added to self.tables.
        """
        tname = table_def['table_name']
        if tname in self.tables:
            raise ValueError('Table %s already exists.' % tname)
        r = catalog.post(
            '%s/table' % self.uri_path,
            json=table_def,
        )
        r.raise_for_status()
        newtable = Table(self.name, tname, r.json())
        self.tables[tname] = newtable
        return newtable

    def delete(self, catalog, model=None):
        """Remove this schema from the remote database.

        Also remove this schema from the local model object (if provided).

        :param catalog: an ErmrestCatalog object
        :param schema: a Schema object or None

        """
        if model is not None:
            if self.name not in model.schemas:
                raise ValueError('Schema %s does not appear to belong to model.' % (self,))
        catalog.delete(self.update_uri_path).raise_for_status()
        if model is not None:
            del model.schemas[self.name]

class Table (_ec.CatalogTable):
    """Named table.
    """
    def __init__(self, sname, tname, table_doc, **kwargs):
        super(Table, self).__init__(sname, tname, table_doc, **_kwargs(**kwargs))
        self.comment = table_doc.get('comment')
        self.kind = table_doc.get('kind')
        self.referenced_by = _ec.MultiKeyedList([])

    @classmethod
    def system_column_defs(cls, custom=[]):
        """Build standard system column definitions, merging optional custom definitions."""
        return [
            Column.define(cname, builtin_types[ctype], nullok)
            for cname, ctype, nullok in [
                    ('RID', 'ermrest_rid', False),
                    ('RCT', 'ermrest_rct', False),
                    ('RMT', 'ermrest_rmt', False),
                    ('RCB', 'ermrest_rcb', True),
                    ('RMB', 'ermrest_rmb', True),
            ]
            if cname not in { c['name']: c for c in custom }
        ] + custom

    @classmethod
    def system_key_defs(cls, custom=[]):
        """Build standard system key definitions, merging optional custom definitions."""
        def ktup(k):
            return tuple(k['unique_columns'])
        return [
            kdef for kdef in [
                Key.define(['RID'])
            ]
            if ktup(kdef) not in { ktup(kdef): kdef for kdef in custom }
        ] + custom

    @classmethod
    def define(cls, tname, column_defs=[], key_defs=[], fkey_defs=[], comment=None, acls={}, acl_bindings={}, annotations={}, provide_system=True):
        """Build a table definition.

        :param tname: the name of the newly defined table
        :param column_defs: a list of Column.define() results for extra or overridden column definitions
        :param key_defs: a list of Key.define() results for extra or overridden key constraint definitions
        :param fkey_defs: a list of ForeignKey.define() results for foreign key definitions
        :param comment: a comment string for the table
        :param acls: a dictionary of ACLs for specific access modes
        :param acl_bindings: a dictionary of dynamic ACL bindings
        :param annotations: a dictionary of annotations
        :param provide_system: whether to inject standard system column definitions when missing from column_defs

        """
        if provide_system:
            column_defs = cls.system_column_defs(column_defs)
            key_defs = cls.system_key_defs(key_defs)

        return {
            'table_name': tname,
            'column_definitions': column_defs,
            'keys': key_defs,
            'foreign_keys': fkey_defs,
            'comment': comment,
            'acls': acls,
            'acl_bindings': acl_bindings,
            'annotations': annotations,
        }

    @classmethod
    def define_vocabulary(cls, tname, curie_template, uri_template='/id/{RID}', column_defs=[], key_defs=[], fkey_defs=[], comment=None, acls={}, acl_bindings={}, annotations={}, provide_system=True):
        """Build a vocabulary table definition.

        :param tname: the name of the newly defined table
        :param curie_template: the RID-based template for the CURIE of locally-defined terms, e.g. 'MYPROJECT:{RID}'
        :param uri_template: the RID-based template for the URI of locally-defined terms, e.g. 'https://server.example.org/id/{RID}'
        :param column_defs: a list of Column.define() results for extra or overridden column definitions
        :param key_defs: a list of Key.define() results for extra or overridden key constraint definitions
        :param fkey_defs: a list of ForeignKey.define() results for foreign key definitions
        :param comment: a comment string for the table
        :param acls: a dictionary of ACLs for specific access modes
        :param acl_bindings: a dictionary of dynamic ACL bindings
        :param annotations: a dictionary of annotations
        :param provide_system: whether to inject standard system column definitions when missing from column_defs

        These core vocabulary columns are generated automatically if
        absent from the input column_defs.

        - ID: ermrest_curie, unique not null, default curie template "%s:{RID}" % curie_prefix
        - URI: ermrest_uri, unique not null, default URI template "/id/{RID}"
        - Name: text, unique not null
        - Description: markdown, not null
        - Synonyms: text[]

        However, caller-supplied definitions override the default.

        """
        if not re.match("^[A-Za-z][-_A-Za-z0-9]*:[{]RID[}]$", curie_template):
            raise ValueError("The curie_template '%s' is invalid." % curie_template)

        if not re.match("^[-_.~?%#=&!*()@:;/+$A-Za-z0-9]+[{]RID[}][-_.~?%#=&!*()@:;/+$A-Za-z0-9]*$", uri_template):
            raise ValueError("The uri_template '%s' is invalid." % uri_template)

        def add_vocab_columns(custom):
            return [
                col_def
                for col_def in [
                        Column.define(
                            'ID',
                            builtin_types['ermrest_curie'],
                            nullok=False,
                            default=curie_template,
                            comment='The preferred Compact URI (CURIE) for this term.'
                        ),
                        Column.define(
                            'URI',
                            builtin_types['ermrest_uri'],
                            nullok=False,
                            default=uri_template,
                            comment='The preferred URI for this term.'
                        ),
                        Column.define(
                            'Name',
                            builtin_types['text'],
                            nullok=False,
                            comment='The preferred human-readable name for this term.'
                        ),
                        Column.define(
                            'Description',
                            builtin_types['markdown'],
                            nullok=False,
                            comment='A longer human-readable description of this term.'
                        ),
                        Column.define(
                            'Synonyms',
                            builtin_types['text[]'],
                            comment='Alternate human-readable names for this term.'
                        ),
                ]
                if col_def['name'] not in { c['name']: c for c in custom }
            ] + custom

        def add_vocab_keys(custom):
            def ktup(k):
                return tuple(k['unique_columns'])
            return [
                key_def
                for key_def in [
                        Key.define(['ID']),
                        Key.define(['URI']),
                ]
                if ktup(key_def) not in { ktup(kdef): kdef for kdef in custom }
            ] + custom

        return cls.define(
            tname,
            add_vocab_columns(column_defs),
            add_vocab_keys(key_defs),
            fkey_defs,
            comment if comment is not None else 'A set of controlled vocabular terms.',
            acls,
            acl_bindings,
            annotations,
            provide_system
        )

    @classmethod
    def define_asset(cls,
                     sname,
                     tname,
                     hatrac_template=None,
                     column_defs=[],
                     key_defs=[],
                     fkey_defs=[],
                     comment=None,
                     acls={},
                     acl_bindings={},
                     annotations={},
                     provide_system=True):
        """Build an asset  table definition.

          :param sname: the name of the schema for the asset table
          :param tname: the name of the newly defined table
          :param hatrac_template: template for the hatrac URL.  Will undergo substitution to template can include
                 elmenents such at {{{MD5}}} or {{{Filename}}}. The default template puts files in
                     /hatrac/schema_name/table_name/md5.filename
                 where the filename and md5 value is computed on upload and the schema_name and table_name are the
                 values of the provided arguments.  If value is set to False, no hatrac_template is used.
          :param column_defs: a list of Column.define() results for extra or overridden column definitions
          :param key_defs: a list of Key.define() results for extra or overridden key constraint definitions
          :param fkey_defs: a list of ForeignKey.define() results for foreign key definitions
          :param comment: a comment string for the table
          :param acls: a dictionary of ACLs for specific access modes
          :param acl_bindings: a dictionary of dynamic ACL bindings
          :param annotations: a dictionary of annotations
          :param provide_system: whether to inject standard system column definitions when missing from column_defs

          These core asset table columns are generated automatically if
          absent from the input column_defs.

          - Filename: ermrest_curie, unique not null, default curie template "%s:{RID}" % curie_prefix
          - URL: Location of the asset, unique not null.  Default template is:
                    /hatrac/sname/tname/{{{MD5}}}.{{{Filename}}} where tname is the name of the asset table.
          - Length: Length of the asset.
          - MD5: text
          - Description: markdown, not null

          However, caller-supplied definitions override the default.

          In addition to creating the columns, this function also creates an asset annotation on the URL column to
          facilitate use of the table by Chaise.
          """

        if hatrac_template is None:
            hatrac_template = '/hatrac/%s/%s/{{{MD5}}}.{{#encode}}{{{Filename}}}{{/encode}}' % (sname, tname)

        def add_asset_annotations(custom):
            annotations.update(custom)
            return annotations

        def add_asset_columns(custom):
            asset_annotation = {
                _ec.tag.asset: {
                    'filename_column': 'Filename',
                    'byte_count_column': 'Length',
                    'md5': 'MD5',
                }
            }
            if hatrac_template:
                asset_annotation[_ec.tag.asset]['url_pattern'] = hatrac_template
            return [
                col_def
                for col_def in [
                        Column.define(
                            'URL', builtin_types['text'],
                            nullok=False,
                            annotations=asset_annotation,
                            comment='URL to the asset',
                        ),
                        Column.define('Filename', builtin_types['text'], comment='Filename of the asset that was uploaded'),
                        Column.define('Description', builtin_types['markdown'], comment='Description of the asset'),
                        Column.define('Length', builtin_types['int8'], nullok=False, comment='Asset length (bytes)'),
                        Column.define('MD5', builtin_types['text'], nullok=False, comment='Asset content MD5 checksum'),
                ]
                if col_def['name'] not in {c['name']: c for c in custom}
            ] + custom

        def add_asset_keys(custom):
            def ktup(k):
                return tuple(k['unique_columns'])

            return [
                key_def
                for key_def in [
                    Key.define(['URL']),
                ]
                if ktup(key_def) not in {ktup(kdef): kdef for kdef in custom}
            ] + custom

        return cls.define(
            tname,
            add_asset_columns(column_defs),
            add_asset_keys(key_defs),
            fkey_defs,
            comment if comment is not None else 'Asset table.',
            acls,
            acl_bindings,
            add_asset_annotations(annotations),
            provide_system
        )

    def prejson(self, prune=True):
        d = super(Table, self).prejson(prune)
        d.update({
            'comment': self.comment,
        })
        return d

    def _create_table_part(self, catalog, subapi, registerfunc, constructor, doc):
        r = catalog.post(
            '%s/%s' % (self.uri_path, subapi),
            json=doc,
        )
        r.raise_for_status()
        created = r.json()
        if isinstance(created, list):
            # handle fkey case where POST returns a list
            assert len(created) == 1
            created = created[0]
        return registerfunc(constructor(self.sname, self.name, created))

    def create_column(self, catalog, column_def):
        """Add a new column to this table in the remote database based on column_def.

           Returns a new Column instance based on the server-supplied
           representation of the new column, and adds it to
           self.column_definitions too.
        """
        cname = column_def['name']
        if cname in self.column_definitions.elements:
            raise ValueError('Column %s already exists.' % cname)
        def add_column(col):
            self.column_definitions.append(col)
            return col
        return self._create_table_part(catalog, 'column', add_column, Column, column_def)

    def create_key(self, catalog, key_def):
        """Add a new key to this table in the remote database based on key_def.

           Returns a new Key instance based on the server-supplied
           representation of the new key, and adds it to self.keys
           too.

        """
        def add_key(key):
            self.keys.append(key)
            return key
        return self._create_table_part(catalog, 'key', add_key, Key, key_def)

    def create_fkey(self, catalog, fkey_def):
        """Add a new foreign key to this table in the remote database based on fkey_def.

           Returns a new ForeignKey instance based on the
           server-supplied representation of the new foreign key, and
           adds it to self.fkeys too.

        """
        def add_fkey(fkey):
            self.foreign_keys.append(fkey)
            return fkey
        return self._create_table_part(catalog, 'foreignkey', add_fkey, ForeignKey, fkey_def)

    def delete(self, catalog, schema=None):
        """Remove this table from the remote database.

        Also remove this table from the local schema object (if provided).

        :param catalog: an ErmrestCatalog object
        :param schema: a Schema object or None

        """
        if schema is not None:
            if self.name not in schema.tables:
                raise ValueError('Table %s does not appear to belong to schema %s.' % (self, schema))
        catalog.delete(self.update_uri_path).raise_for_status()
        if schema is not None:
            del schema.tables[self.name]

class Column (_ec.CatalogColumn):
    """Named column.
    """
    def __init__(self, sname, tname, column_doc, **kwargs):
        super(Column, self).__init__(sname, tname, column_doc, **_kwargs(**kwargs))
        self.type = make_type(column_doc['type'], **_kwargs(**kwargs))
        self.nullok = bool(column_doc.get('nullok', True))
        self.default = column_doc.get('default')
        self.comment = column_doc.get('comment')

    @classmethod
    def define(cls, cname, ctype, nullok=True, default=None, comment=None, acls={}, acl_bindings={}, annotations={}):
        """Build a column definition."""
        if not isinstance(ctype, Type):
            raise TypeError('Ctype %s should be an instance of Type.' % ctype)
        if not isinstance(nullok, bool):
            raise TypeError('Nullok %s should be an instance of bool.' % nullok)
        return {
            'name': cname,
            'type': ctype.prejson(),
            'nullok': nullok,
            'default': default,
            'comment': comment,
            'acls': acls,
            'acl_bindings': acl_bindings,
            'annotations': annotations,
        }

    def prejson(self, prune=True):
        d = super(Column, self).prejson(prune)
        d.update({
            'type': self.type.prejson(prune),
            'nullok': self.nullok,
            'default': self.default,
            'comment': self.comment,
        })
        return d
        
    def delete(self, catalog, table=None):
        """Remove this column from the remote database.

        Also remove this column from the local table object (if provided).

        :param catalog: an ErmrestCatalog object
        :param table: a Table object or None

        """
        if table is not None:
            if self.name not in table.column_definitions.elements:
                raise ValueError('Column %s does not appear to belong to table %s.' % (self, table))
        catalog.delete(self.update_uri_path).raise_for_status()
        if table is not None:
            del table.column_definitions[self.name]

class Key (_ec.CatalogKey):
    """Named key.
    """
    def __init__(self, sname, tname, key_doc, **kwargs):
        super(Key, self).__init__(sname, tname, key_doc, **_kwargs(**kwargs))
        self.comment = key_doc.get('comment')

    @classmethod
    def define(cls, colnames, constraint_names=[], comment=None, annotations={}):
        """Build a key definition."""
        if not isinstance(colnames, list):
            raise TypeError('Colnames should be a list.')
        return {
            'unique_columns': list(colnames),
            'names': constraint_names,
            'comment': comment,
            'annotations': annotations,
        }

    def prejson(self, prune=True):
        d = super(Key, self).prejson(prune)
        d.update({
            'comment': self.comment,
        })
        return d

    def delete(self, catalog, table=None):
        """Remove this key from the remote database.

        Also remove this key from the local table object (if provided).

        :param catalog: an ErmrestCatalog object
        :param table: a Table object or None

        """
        if table is not None:
            if self.names[0] not in table.keys.elements:
                raise ValueError('Key %s does not appear to belong to table %s.' % (self, table))
        catalog.delete(self.update_uri_path).raise_for_status()
        if table is not None:
            del table.keys[self.names[0]]

class ForeignKey (_ec.CatalogForeignKey):
    """Named foreign key.
    """
    def __init__(self, sname, tname, fkey_doc, **kwargs):
        super(ForeignKey, self).__init__(sname, tname, fkey_doc, **_kwargs(**kwargs))
        self.comment = fkey_doc.get('comment')
        self.on_delete = fkey_doc.get('on_delete')
        self.on_update = fkey_doc.get('on_update')

    @classmethod
    def define(cls, fk_colnames, pk_sname, pk_tname, pk_colnames, on_update='NO ACTION', on_delete='NO ACTION', constraint_names=[], comment=None, acls={}, acl_bindings={}, annotations={}):
        if len(fk_colnames) != len(pk_colnames):
            raise ValueError('The fk_colnames and pk_colnames lists must have the same length.')
        return {
            'foreign_key_columns': [
                {
                    'column_name': fk_colname
                }
                for fk_colname in fk_colnames
            ],
            'referenced_columns': [
                {
                    'schema_name': pk_sname,
                    'table_name': pk_tname,
                    'column_name': pk_colname,
                }
                for pk_colname in pk_colnames
            ],
            'on_update': on_update,
            'on_delete': on_delete,
            'names': constraint_names,
            'comment': comment,
            'acls': acls,
            'acl_bindings': acl_bindings,
            'annotations': annotations,
        }

    def prejson(self, prune=True):
        d = super(ForeignKey, self).prejson(prune)
        d.update({
            'comment': self.comment,
            'on_delete': self.on_delete,
            'on_update': self.on_update,
        })
        return d

    def delete(self, catalog, table=None):
        """Remove this foreign key from the remote database.

        Also remove this foreign key from the local table object (if provided).

        :param catalog: an ErmrestCatalog object
        :param table: a Table object or None

        """
        if table is not None:
            if self.names[0] not in table.foreign_keys.elements:
                raise ValueError('Foreign key %s does not appear to belong to table %s.' % (self, table))
        catalog.delete(self.update_uri_path).raise_for_status()
        if table is not None:
            del table.foreign_keys[self.names[0]]

def make_type(type_doc, **kwargs):
    """Create instance of Type, DomainType, or ArrayType as appropriate for type_doc."""
    if type_doc.get('is_domain', False):
        return DomainType(type_doc, **kwargs)
    elif type_doc.get('is_array', False):
        return ArrayType(type_doc, **kwargs)
    else:
        return Type(type_doc, **kwargs)

class Type (object):
    """Named type.
    """
    def __init__(self, type_doc, **kwargs):
        self.typename = type_doc['typename']
        self.is_domain = False
        self.is_array = False

    def prejson(self, prune=True):
        d = {
            'typename': self.typename,
        }
        return d

class DomainType (Type):
    """Named domain type.
    """
    def __init__(self, type_doc, **kwargs):
        super(DomainType, self).__init__(type_doc, **kwargs)
        self.is_domain = True
        self.base_type = make_type(type_doc['base_type'], **kwargs)
        
    def prejson(self, prune=True):
        d = super(DomainType, self).prejson(prune)
        d.update({
            'is_domain': True,
            'base_type': self.base_type.prejson(prune)
        })
        return d

class ArrayType (Type):
    """Named domain type.
    """
    def __init__(self, type_doc, **kwargs):
        super(ArrayType, self).__init__(type_doc, **kwargs)
        is_array = True
        self.base_type = make_type(type_doc['base_type'], **kwargs)

    def prejson(self, prune=True):
        d = super(ArrayType, self).prejson(prune)
        d.update({
            'is_array': True,
            'base_type': self.base_type.prejson(prune)
        })
        return d

builtin_types = _ec.AttrDict(
    # first define standard scalar types
    {
        typename: Type({'typename': typename})
        for typename in {
                'date',
                'float4', 'float8',
                'json', 'jsonb',
                'int2', 'int4', 'int8',
                'text',
                'timestamptz', 'timestamp',
                'boolean'
        }
    }
)
builtin_types.update(
    # define some typical array types
    {
        '%s[]' % typename: ArrayType({
            'typename': '%s[]' % typename,
            'is_array': True,
            'base_type': typedoc.prejson()
        })
        for typename, typedoc in builtin_types.items()
    }
)
builtin_types.update(
    # define standard domain types
    {
        domain: DomainType({
            'typename': domain,
            'is_domain': True,
            'base_type': builtin_types[basetypename].prejson(),
        })
        for domain, basetypename in {
                'ermrest_rid': 'text',
                'ermrest_rcb': 'text',
                'ermrest_rmb': 'text',
                'ermrest_rct': 'timestamptz',
                'ermrest_rmt': 'timestamptz',
                'markdown': 'text',
                'ermrest_curie': 'text',
                'ermrest_uri': 'text',
        }.items()
    }
)
builtin_types.update(
    # define standard serial types which don't have array types
    {
        typename: Type({'typename': typename})
        for typename in [ 'serial2', 'serial4', 'serial8' ]
    }
)




from . import urlquote


class AttrDict (dict):
    """Dictionary with optional attribute-based lookup.

       For keys that are valid attributes, self.key is equivalent to
       self[key].
    """
    def __getattr__(self, a):
        try:
            return self[a]
        except KeyError as e:
            raise AttributeError(str(e))

    def __setattr__(self, a, v):
        self[a] = v

# convenient enumeration of common annotation tags
tag = AttrDict({
    'generated':          'tag:isrd.isi.edu,2016:generated',
    'immutable':          'tag:isrd.isi.edu,2016:immutable',
    'display':            'tag:misd.isi.edu,2015:display',
    'visible_columns':    'tag:isrd.isi.edu,2016:visible-columns',
    'visible_foreign_keys': 'tag:isrd.isi.edu,2016:visible-foreign-keys',
    'foreign_key':        'tag:isrd.isi.edu,2016:foreign-key',
    'table_display':      'tag:isrd.isi.edu,2016:table-display',
    'table_alternatives': 'tag:isrd.isi.edu,2016:table-alternatives',
    'column_display':     'tag:isrd.isi.edu,2016:column-display',
    'asset':              'tag:isrd.isi.edu,2017:asset',
    'bulk_upload':        'tag:isrd.isi.edu,2017:bulk-upload',
    'export':             'tag:isrd.isi.edu,2016:export',
})

def presence_annotation(tag_uri):
    """Decorator to establish property getter/setter/deleter for presence annotations.

       Usage example:

          @presence_annotation(tag.generated)
          def generated(self): pass

       The stub method will be discarded.
    """
    def helper(ignore):
        docstr = "Convenience property for managing presence of annotation %s" % tag_uri

        def getter(self):
            return tag_uri in self.annotations

        def setter(self, present):
            if present:
                self.annotations[tag_uri] = None
            else:
                self.annotations.pop(tag_uri, None)

        def deleter(self):
            self.annotations.pop(tag_uri, None)

        return property(getter, setter, deleter, docstr)

    return helper

def object_annotation(tag_uri):
    """Decorator to establish property getter/setter/deleter for object annotations.

       Usage example:

          @presence_annotation(tag.display)
          def display(self): pass

       The stub method will be discarded.
    """
    def helper(ignore):
        docstr = "Convenience property for managing content of object annotation %s" % tag_uri

        def getter(self):
            if tag_uri not in self.annotations:
                self.annotations[tag_uri] = AttrDict({})
            return self.annotations[tag_uri]

        def setter(self, value):
            if not isinstance(value, (dict, AttrDict)):
                raise TypeError('Unexpected object type %s for annotation %s' % (type(value), tag_uri))
            self.annotations[tag_uri] = AttrDict(value)

        def deleter(self):
            self.annotations.pop(tag_uri, None)

        return property(getter, setter, deleter, docstr)

    return helper

def equivalent(doc1, doc2, method=None):
    """Determine whether two dict/array/literal documents are structurally equivalent."""
    if method == 'acl_binding':
        # fill in defaults to avoid some false negatives on acl binding comparison
        if not isinstance(doc1, dict):
            return False
        def canonicalize(d):
            if not isinstance(d, dict):
                return d
            def helper(b):
                if not isinstance(b, dict):
                    return b
                return {
                    'projection': b['projection'],
                    'projection_type': b.get('projection_type'), # we can't provide default w/o type inference!
                    'types': b['types'],
                    'scope_acl': b.get('scope_acl', ['*']), # this is a common omission...
                }
            return {
                binding_name: helper(binding)
                for binding_name, binding in d.items()
            }
        return equivalent(canonicalize(doc1), canonicalize(doc2))
    elif isinstance(doc1, dict) and isinstance(doc2, dict):
        return equivalent(sorted(doc1.items()), sorted(doc2.items()))
    elif isinstance(doc1, (list, tuple)) and isinstance(doc2, (list, tuple)):
        if len(doc1) != len(doc2):
            return False
        for e1, e2 in zip(doc1, doc2):
            if not equivalent(e1, e2):
                return False
        return True
    return doc1 == doc2

class NodeConfig (object):
    """Generic model document node configuration management.

       annotations: map of annotations for node by key
       comment: comment string or None (if supported by sub-class)

       Convenience access for common annotations:
         self.display: access mutable tag.display object
         self.generated: treat tag.generated as a boolean
         self.immutable: treat tag.immutable as a boolean
    """
    def __init__(self, uri_path, node_doc):
        self.uri_path = uri_path
        self.update_uri_path = uri_path
        self.annotations = dict(node_doc.get('annotations', {}))
        self._supports_comment = True
        self._comment = node_doc.get('comment')

    def apply(self, catalog, existing=None):
        """Apply configuration to corresponding node in catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self, or None to apply configuration unconditionally.

        The configuration in self.comment and self.annotations will be
        applied to the remote model node corresponding to self, unless
        existing node configuration is supplied and is equivalent.
        """
        if self._supports_comment:
            if existing is None or not equivalent(self._comment, existing._comment):
                if self._comment is not None:
                    catalog.put('%s/comment' % self.update_uri_path, data=self._comment)
                else:
                    catalog.delete('%s/comment' % self.update_uri_path)
        if existing is None or not equivalent(self.annotations, existing.annotations):
            catalog.put(
                '%s/%s' % (self.update_uri_path, 'annotation'),
                json=self.annotations
            )

    def clear(self, clear_comment=False):
        """Clear existing annotations on node, also clearing comment if clear_comment is True.

        NOTE: as a backwards-compatible heuristic, comments are
        retained by default so that a typical configuration-management
        client does not strip useful documentation from existing models.

        """
        self.annotations.clear()
        if clear_comment and self._supports_comment:
            self.comment = None

    def prejson(self):
        """Produce a representation of configuration as generic Python data structures"""
        d = dict()
        if self.annotations:
            d["annotations"] = self.annotations
        return d

    @property
    def comment(self):
        """Comment on this node in model, if supported.

        Raises TypeError if accessed when unsupported, e.g. on top-level catalog objects.
        """
        if self._supports_comment:
            return self._comment
        else:
            raise TypeError('%r does not support comment management.' % type(self).__name__)

    @comment.setter
    def comment(self, value):
        """Comment on this node in model, if supported.

        Raises TypeError if accessed when unsupported, e.g. on top-level catalog objects.
        """
        if self._supports_comment:
            self._comment = value
        else:
            raise TypeError('%r does not support comment management.' % type(self).__name__)

    @presence_annotation(tag.immutable)
    def immutable(self): pass

    @presence_annotation(tag.generated)
    def generated(self): pass

    @object_annotation(tag.display)
    def display(self): pass

class NodeConfigAcl (NodeConfig):
    """Generic model acl-bearing document node configuration management.

       acls: map of acls for node by key
       annotations: map of annotations for node by key
       comment: comment string or None (if supported by sub-class)

       Convenience access for common annotations:
         self.display: access mutable tag.display object
         self.generated: treat tag.generated as a boolean
         self.immutable: treat tag.immutable as a boolean
    """
    def __init__(self, uri_path, node_doc):
        NodeConfig.__init__(self, uri_path, node_doc)
        self.acls = AttrDict(node_doc.get('acls', {}))

    def apply(self, catalog, existing=None):
        """Apply configuration to corresponding node in catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self, or None to apply configuration unconditionally.

        The configuration in self.comment, self.annotations, and
        self.acls will be applied to the remote model node
        corresponding to self, unless existing node configuration is
        supplied and is equivalent.

        """
        NodeConfig.apply(self, catalog, existing)
        if existing is None or not equivalent(self.acls, existing.acls):
            catalog.put(
                '%s/%s' % (self.update_uri_path, 'acl'),
                json=self.acls
            )

    def clear(self):
        """Clear existing acls and annotations on node."""
        NodeConfig.clear(self)
        self.acls.clear()

    def prejson(self):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfig.prejson(self)
        if self.acls:
            d["acls"] = self.acls
        return d


class NodeConfigAclBinding (NodeConfigAcl):
    """Generic model acl_binding-bearing document node configuration management.

       acl_bindings: map of acl bindings for node by key
       acls: map of acls for node by key
       annotations: map of annotations for node by key
       comment: comment string or None (if supported by sub-class)

       Convenience access for common annotations:
         self.display: access mutable tag.display object
         self.generated: treat tag.generated as a boolean
         self.immutable: treat tag.immutable as a boolean
    """
    def __init__(self, uri_path, node_doc):
        NodeConfigAcl.__init__(self, uri_path, node_doc)
        self.acl_bindings = AttrDict(node_doc.get('acl_bindings', {}))

    def apply(self, catalog, existing=None):
        """Apply configuration to corresponding node in catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self, or None to apply configuration unconditionally.

        The configuration in self.comment, self.annotations,
        self.acls, and self.acl_bindings will be applied to the remote
        model node corresponding to self, unless existing node
        configuration is supplied and is equivalent.

        """
        NodeConfigAcl.apply(self, catalog, existing)
        if existing is None or not equivalent(self.acl_bindings, existing.acl_bindings, method='acl_binding'):
            catalog.put(
                '%s/%s' % (self.update_uri_path, 'acl_binding'),
                json=self.acl_bindings
            )

    def clear(self):
        """Clear existing acl_bindings, acls, and annotations on node."""
        NodeConfigAcl.clear(self)
        self.acl_bindings.clear()

    def prejson(self):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfigAcl.prejson(self)
        if self.acl_bindings:
            d["acl_bindings"] = self.acl_bindings
        return d


class CatalogConfig (NodeConfigAcl):
    """Top-level catalog configuration management.

       acls: catalog-level ACL configuration
       annotations: catalog-level annotations
       schemas: all schemas in catalog, by name
    """
    def __init__(self, model_doc, **kwargs):
        NodeConfigAcl.__init__(self, "/schema", model_doc)
        self._supports_comment = False
        self.update_uri_path = ""
        self.schemas = {
            sname: kwargs.get('schema_class', CatalogSchema)(sname, sdoc, **kwargs)
            for sname, sdoc in model_doc.get('schemas', {}).items()
        }

    @classmethod
    def fromcatalog(cls, catalog):
        """Retrieve catalog config as a CatalogConfig management object."""
        return cls(catalog.get("/schema").json())

    def apply(self, catalog, existing=None):
        """Apply catalog configuration to catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self.

        The configuration in self will be applied recursively to the
        corresponding model nodes in schema. For each node, the
        comment, annotations, acls, and/or acl_bindings will be
        applied where applicable.

        If existing is not provided (default), the current whole
        configuration will be retrieved from the catalog and used
        automatically to determine whether the configuration goals
        under this CatalogConfig instance are already met or need to
        be remotely applied.

        """
        if existing is None:
            existing = self.fromcatalog(catalog)
        NodeConfigAcl.apply(self, catalog, existing)
        for sname, schema in self.schemas.items():
            schema.apply(catalog, existing.schemas[sname])

    def clear(self):
        """Clear all configuration in catalog and children."""
        NodeConfigAcl.clear(self)
        for schema in self.schemas.values():
            schema.clear()

    def table(self, sname, tname):
        """Return table configuration for table with given name."""
        return self.schemas[sname].tables[tname]

    def column(self, sname, tname, cname):
        """Return column configuration for column with given name."""
        return self.table(sname, tname).column_definitions[cname]

    @object_annotation(tag.bulk_upload)
    def bulk_upload(self): pass

    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfigAcl.prejson(self)
        d["schemas"] = {
            sname: schema.prejson()
            for sname, schema in self.schemas.items()
        }
        return d

class CatalogSchema (NodeConfigAcl):
    """Schema-level configuration management.

       acls: schema-level ACL configuration
       annotations: schema-level annotations
       tables: all tables in schema, by name

       Convenience access for common annotations:
         self.display: access mutable tag.display object
    """
    def __init__(self, sname, schema_doc, **kwargs):
        NodeConfigAcl.__init__(
            self,
            "/schema/%s" % urlquote(sname),
            schema_doc
        )
        self.name = sname
        self.tables = {
            tname: kwargs.get('table_class', CatalogTable)(sname, tname, tdoc, **kwargs)
            for tname, tdoc in schema_doc.get('tables', {}).items()
        }

    def apply(self, catalog, existing=None):
        """Apply schema configuration to catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self.

        The configuration in self will be applied recursively to the
        corresponding model nodes in catalog. For each node, the
        comment, annotations, acls, and/or acl_bindings will be
        applied where applicable unless existing value is equivalent.

        """
        NodeConfigAcl.apply(self, catalog, existing)
        for tname, table in self.tables.items():
            table.apply(catalog, existing.tables[tname] if existing else None)

    def clear(self):
        """Clear all configuration in schema and children."""
        NodeConfigAcl.clear(self)
        for table in self.tables.values():
            table.clear()

    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfigAcl.prejson(self)
        d["schema_name"] = self.name
        d["tables"] = {
            tname: table.prejson()
            for tname, table in self.tables.items()
        }
        return d


class KeyedList (list):
    """Keyed list."""
    def __init__(self, l):
        list.__init__(self, l)
        self.elements = {
            e.name: e
            for e in l
        }

    def __getitem__(self, idx):
        """Get element by key or by list index or slice."""
        if isinstance(idx, (int, slice)):
            return list.__getitem__(self, idx)
        else:
            return self.elements[idx]

    def __delitem__(self, idx):
        """Delete element by key or by list index or slice."""
        if isinstance(idx, int):
            victim = list.__getitem__(self, idx)
            list.__delitem__(self, idx)
            del self.elements[victim.name]
        elif isinstance(idx, slice):
            victims = [list.__getitem__(self, idx)]
            list.__delslice__(self, idx)
            for victim in victims:
                del self.elements[victim.name]
        else:
            victim = self.elements[idx]
            list.__delitem__(self, self.index(victim))
            del self.elements[victim.name]

    def append(self, e):
        """Append element to list and record its key."""
        if e.name in self.elements:
            raise ValueError('Element name %s already exists.' % e.name)
        list.append(self, e)
        self.elements[e.name] = e

class MultiKeyedList (list):
    """Multi-keyed list."""
    def __init__(self, l):
        list.__init__(self, l)
        self.elements = {
            tuple(name): e
            for e in l
            for name in e.names
        }

    def __getitem__(self, idx):
        """Get element by key or by list index or slice."""
        if isinstance(idx, (tuple, list)):
            return self.elements[idx]
        else:
            return list.__getitem__(self, idx)

    def __delitem__(self, idx):
        """Delete element by key or by list index."""
        item = self[idx]
        list.__delitem__(self, self.index(item))
        for name in item.names:
            del self.elements[name]

    def append(self, e):
        """Append element to list and record its keys."""
        for name in e.names:
            if name in self.elements:
                raise ValueError('Element name %s already exists.' % e.names)
        list.append(self, e)
        for name in e.names:
            self.elements[name] = e

class CatalogTable (NodeConfigAclBinding):
    """Table-level configuration management.

       acl_bindings: table-level dynamic ACL bindings
       acls: table-level ACL configuration
       annotations: table-level annotations
       column_definitions: columns in table

       Convenience access to common annotations:
         self.alternatives: tag.table_alternatives object
         self.display: tag.display object
         self.generated: treat tag.generated as a boolean
         self.immutable: treat tag.immutable as a boolean
         self.table_display: tag.table_display object
         self.visible_columns: tag.visible_columns object
         self.visible_foreign_keys: tag.visible_foreign_keys object
    """

    def __init__(self, sname, tname, table_doc, **kwargs):
        NodeConfigAclBinding.__init__(
            self,
            "/schema/%s/table/%s" % (urlquote(sname), urlquote(tname)),
            table_doc
        )
        self.sname = sname
        self.name = tname
        self.column_definitions = KeyedList([
            kwargs.get('column_class', CatalogColumn)(sname, tname, cdoc, **kwargs)
            for cdoc in table_doc.get('column_definitions', [])
        ])
        self.keys = MultiKeyedList([
            kwargs.get('key_class', CatalogKey)(sname, tname, kdoc, **kwargs)
            for kdoc in table_doc.get('keys', [])
        ])
        self.foreign_keys = MultiKeyedList([
            kwargs.get('foreign_key_class', CatalogForeignKey)(sname, tname, fkdoc, **kwargs)
            for fkdoc in table_doc.get('foreign_keys', [])
        ])

    def apply(self, catalog, existing=None):
        """Apply table configuration to catalog unless existing already matches.

        :param catalog: The EmrestCatalog instance to which configuration will be applied.
        :param existing: An instance comparable to self.

        The configuration in self will be applied recursively to the
        corresponding model nodes in catalog. For each node, the
        comment, annotations, acls, and/or acl_bindings will be
        applied where applicable unless existing is supplied and is
        equivalent.

        """
        NodeConfigAclBinding.apply(self, catalog, existing)
        for col in self.column_definitions:
            col.apply(catalog, existing.column_definitions[col.name] if existing else None)
        for key in self.keys:
            key.apply(catalog, existing.keys[key.names[0]] if existing else None)
        for fkey in self.foreign_keys:
            fkey.apply(catalog, existing.foreign_keys[fkey.names[0]] if existing else None)

    def clear(self):
        """Clear all configuration in table and children."""
        NodeConfigAclBinding.clear(self)
        for col in self.column_definitions:
            col.clear()
        for key in self.keys:
            key.clear()
        for fkey in self.foreign_keys:
            fkey.clear()

    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfigAclBinding.prejson(self)
        d["schema_name"] = self.sname
        d["table_name"] = self.name
        d["column_definitions"] = [
            column.prejson()
            for column in self.column_definitions
        ]
        d["keys"] = [
            key.prejson()
            for key in self.keys
        ]
        d["foreign_keys"] = [
            fkey.prejson()
            for fkey in self.foreign_keys
        ]
        return d

    @object_annotation(tag.table_alternatives)
    def alternatives(self): pass

    @object_annotation(tag.table_display)
    def table_display(self): pass

    @object_annotation(tag.visible_columns)
    def visible_columns(self): pass

    @object_annotation(tag.visible_foreign_keys)
    def visible_foreign_keys(self): pass

class CatalogColumn (NodeConfigAclBinding):
    """Column-level configuration management.

       acl_bindings: column-level dynamic ACL bindings
       acls: column-level ACL configuration
       annotations: column-level annotations
       name: name of column

       Convenience access to common annotations:
         self.asset: tag.asset object
         self.column_display:: tag.column_display object
         self.display: tag.display object
         self.generated: treat tag.generated as a boolean
         self.immutable: treat tag.immutable as a boolean
    """

    def __init__(self, sname, tname, column_doc, **kwargs):
        cname = column_doc['name']
        NodeConfigAclBinding.__init__(
            self,
            "/schema/%s/table/%s/column/%s" % (urlquote(sname), urlquote(tname), urlquote(cname)),
            column_doc
        )
        self.sname = sname
        self.tname = tname
        self.name = cname

    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfig.prejson(self)
        d["name"] = self.name
        return d

    @object_annotation(tag.asset)
    def asset(self): pass

    @object_annotation(tag.column_display)
    def column_display(self): pass

class CatalogKey (NodeConfig):
    """Key-level configuration management.

       annotations: column-level annotations
       names: name(s) of key constraint
    """
    def __init__(self, sname, tname, key_doc, **kwargs):
        NodeConfig.__init__(
            self,
            '/schema/%s/table/%s/key/%s' % (
                urlquote(sname),
                urlquote(tname),
                ','.join([ urlquote(cname) for cname in key_doc['unique_columns'] ])
            ),
            key_doc
        )
        self.sname = sname
        self.tname = tname
        self.names = [ tuple(name) for name in key_doc['names'] ]
        self.unique_columns = key_doc['unique_columns']

    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        d = NodeConfig.prejson(self)
        d['unique_columns'] = self.unique_columns
        d['names'] = self.names
        return d


class CatalogForeignKey (NodeConfigAclBinding):
    """Foreign key-level configuration management.

       acl_bindings: foreign key-level acl-bindings
       acls: foreign key-level acls
       annotations: foreign key-level annotations
    """
    def __init__(self, sname, tname, fkey_doc, **kwargs):
        refcols = fkey_doc['referenced_columns']
        NodeConfigAclBinding.__init__(
            self,
            '/schema/%s/table/%s/foreignkey/%s/reference/%s:%s/%s' % (
                urlquote(sname),
                urlquote(tname),
                ','.join([ urlquote(col['column_name']) for col in fkey_doc['foreign_key_columns'] ]),
                urlquote(refcols[0]['schema_name']),
                urlquote(refcols[0]['table_name']),
                ','.join([ urlquote(col['column_name']) for col in refcols ]),
            ),
            fkey_doc
        )
        self.sname = sname
        self.tname = tname
        self.names = [ tuple(name) for name in fkey_doc['names'] ]
        self.foreign_key_columns = fkey_doc['foreign_key_columns']
        self.referenced_columns = fkey_doc['referenced_columns']
    
    def prejson(self, prune=True):
        """Produce a representation of configuration as generic Python data structures"""
        def expand(c):
            c['schema_name'] = self.sname
            c['table_name'] = self.tname
            return c
        d = NodeConfig.prejson(self)
        d['foreign_key_columns'] = [
            expand(c)
            for c in self.foreign_key_columns
        ]
        d['referenced_columns'] = self.referenced_columns
        d['names'] = self.names
        return d

    @object_annotation(tag.foreign_key)
    def foreign_key(self): pass



import os
import sys
import platform
import argparse
import traceback
from deriva.core import DerivaServer, ErmrestCatalog, HatracStore, AttrDict, BaseCLI, KeyValuePairArgs, \
    get_credential, format_credential, __version__
from deriva.core.ermrest_model import builtin_types, Table, Column, Key, ForeignKey


class DerivaInitializer(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.hostname = self.kwargs.get("hostname", platform.node())
        self.catalog_id = self.kwargs.get("catalog_id", 1)

        self.server = None
        self.catalog = None
        self.object_store = None
        self.object_store_namespace = "/hatrac/"
        self.groups = AttrDict({})

        # get credential
        token = kwargs.get("token")
        credential_file = kwargs.get("credential_file")
        if token:
            self.credentials = format_credential(token=token)
        elif credential_file:
            self.credentials = get_credential(self.hostname, credential_file)
        else:
            self.credentials = get_credential(self.hostname)


class TutorialInitializer(DerivaInitializer):

    def __init__(self, **kwargs):
        super(TutorialInitializer, self).__init__(**kwargs)

    def initialize_groups(self):
        # initialize groups
        self.groups = AttrDict({
            "admin":   "https://auth.globus.org/5a773142-e2ed-11e8-a017-0e8017bdda58",
            "curator": "https://auth.globus.org/a5cfa412-e2ed-11e8-a768-0e368f3075e8",
            "writer":  "https://auth.globus.org/caa11064-e2ed-11e8-9d6d-0a7c1eab007a",
            "reader":  "https://auth.globus.org/b9100ea4-e2ed-11e8-8b39-0e368f3075e8",
        })

    def initialize_hatrac(self):
        self.object_store = HatracStore('https', self.hostname, self.credentials)
        self.object_store_namespace += "tutorial_project_data"
        if not self.object_store.is_valid_namespace(self.object_store_namespace):
            self.object_store.create_namespace(self.object_store_namespace)
        self.object_store.set_acl(self.object_store_namespace, 'owner', [self.groups.admin])
        self.object_store.set_acl(self.object_store_namespace, 'subtree-create', [self.groups.curator, self.groups.writer])
        self.object_store.set_acl(self.object_store_namespace, 'subtree-update', [self.groups.curator])
        self.object_store.set_acl(self.object_store_namespace, 'subtree-read', [self.groups.reader])

    def initialize_catalog(self):
        self.server = DerivaServer('https', self.hostname, self.credentials)
        self.catalog = self.server.connect_ermrest(self.catalog_id)
        if not self.catalog.exists():
            self.catalog = self.server.create_ermrest_catalog()
        model = self.catalog.getCatalogModel()

        # 1. Modify local representation of catalog ACL config
        model.acls.update({
            "owner": [self.groups.admin],
            "insert": [self.groups.curator, self.groups.writer],
            "update": [self.groups.curator],
            "delete": [self.groups.curator],
            "select": [self.groups.writer, self.groups.reader],
            "enumerate": ["*"]
        })
        # apply these local config changes to the server
        model.apply(self.catalog)

        # 2. Mutate local configuration of client_table which is part of model
        client_table = model.schemas["public"].tables["ermrest_client"]
        client_table.acls.update({
          "select": [self.groups.curator, self.groups.writer, self.groups.reader],
        })
        client_table.column_definitions["client_obj"].acls.update({
          "select": [],
        })
        # apply these local changes to server
        model.apply(self.catalog)

        # 3. Create "Journal" table
        model.schemas["public"].create_table(
            Table.define(
                "Journal",
                [
                    Column.define(
                        "Notes",
                        builtin_types["markdown"],
                        nullok=False,
                        comment="User-provided notes.",
                    )
                ],
                comment="A journal of user-provided notes."
            )
        )
        # retrieve catalog model again to ensure we reflect latest structural changes
        model = self.catalog.getCurrentModel()

        # 4. Create "Journal_Attachment" asset table
        model.schemas["public"].create_table(
            Table.define(
                "Journal_Attachment",
                [
                    Column.define(
                        "journal_rid",
                        builtin_types["text"],
                        nullok=False,
                        comment="The journal entry to which this asset is attached."
                    ),
                    Column.define(
                        "url",
                        builtin_types["text"],
                        nullok=False,
                        comment="The URL of the stored attachment."
                    ),
                    Column.define(
                        "length",
                        builtin_types["int8"],
                        nullok=False,
                        comment="The asset length (byte count)."
                    ),
                    Column.define(
                        "md5",
                        builtin_types["text"],
                        nullok=False,
                        comment="The hexadecimal encoded MD5 checksum of the asset."
                    ),
                    Column.define(
                        "content_type",
                        builtin_types["text"],
                        nullok=True,
                        comment="The content-type of the asset."
                    ),
                    Column.define(
                        "file_name",
                        builtin_types["text"],
                        nullok=True,
                        comment="The suggested local filename on client systems."
                    ),
                ],
                fkey_defs=[
                    ForeignKey.define(
                        ["journal_rid"],
                        "public",
                        "Journal",
                        ["RID"],
                        on_delete="CASCADE",
                        constraint_names=[["public", "Journal_Attachment_journal_rid_fkey"]]
                    ),
                ],
                comment="Assets (files) attached to Journal entries.",
            )
        )
        # retrieve catalog model again to ensure we reflect latest structural changes
        model = self.catalog.getCurrentModel()

        # 5. Annotate "Journal_Attachment" table
        attachment_table = model.schemas["public"].tables["Journal_Attachment"]
        url_column = attachment_table.column_definitions["url"]
        url_column.annotations.update({
            "tag:isrd.isi.edu,2017:asset": {
                "filename_column": "file_name",
                "byte_count_column": "length",
                "md5": "md5",
                "url_pattern": "/hatrac/%s/journal_attachment/{{{journal_rid}}}/{{{_url.md5_hex}}}" %
                               self.object_store_namespace
            }
        })
        attachment_table.annotations.update({
            "tag:isrd.isi.edu,2015:display": {
                "name_style": {"underline_space": True}
            },
            "tag:isrd.isi.edu,2016:visible-columns": {
                "entry": [
                    ["public", "Journal_Attachment_journal_rid_fkey"],
                    "url"
                ]
            }
        })
        model.apply(self.catalog)

        # 6. Allow Self-Service Editing by Writers
        journal_table = model.schemas["public"].tables["Journal"]
        attachment_table = model.schemas["public"].tables["Journal"]
        # we can re-use this generic policy on any table since they all have an RCB column
        self_service_policy = {
            "self_service": {
                "types": ["update", "delete"],
                "projection": ["RCB"],
                "projection_type": "acl"
            }
        }
        journal_table.acl_bindings.update(self_service_policy)
        attachment_table.acl_bindings.update(self_service_policy)

        # apply these local config changes to the server
        model.apply(self.catalog)

        # 7. Prevent Attachments by Unrelated Writers
        journal_fkey = attachment_table.foreign_keys[
            ("public", "Journal_Attachment_journal_rid_fkey")
        ]
        journal_fkey.acls.update({
            "insert": [grps.curator],
            "update": [grps.curator],
        })
        journal_fkey.acl_bindings.update({
            "self_linkage": {
                "types": ["insert", "update"],
                "projection": ["RCB"],
                "projection_type": "acl",
            }
        })

        # 8. Add column to "Journal" table
        # retrieve catalog model again to ensure we reflect latest structural changes
        model = self.catalog.getCurrentModel()
        journal_table.add_column(
            "Date",
            builtin_types["timestamptz"],
            nullok=False,
            comment="User-provided timestamp."
        )
        # apply these local config changes to the server
        model.apply(self.catalog)

    def initialize(self):
        self.initialize_groups()
        self.initialize_hatrac()
        self.initialize_catalog()


class DerivaTutorialCLI(BaseCLI):
    def __init__(self, description, epilog):

        BaseCLI.__init__(self, description, epilog, __version__)
        self.remove_options(['--host', '--config-file'])
        self.parser.add_argument('host', default='localhost', metavar='<host>', help="Fully qualified host name.")
        self.parser.add_argument("kwargs", metavar="[key=value key=value ...]",
                                 nargs=argparse.REMAINDER, action=KeyValuePairArgs, default=dict(),
                                 help="Variable length of whitespace-delimited key=value pair arguments. "
                                      "For example: key1=value1 key2=value2")

    def main(self):
        try:
            args = self.parse_cli()
        except ValueError as e:
            sys.stderr.write(str(e))
            return 2
        if not args.quiet:
            sys.stderr.write("\n")

        try:
            tutorial = TutorialInitializer(hostname=args.host,
                                           token=args.token,
                                           credential_file=args.credential_file,
                                           **args.kwargs)
            tutorial.initialize()
        except:
            traceback.print_exc()
            return 1
        finally:
            if not args.quiet:
                sys.stderr.write("\n\n")
        return 0


def main():
    desc = "DERIVA Tutorial Command-Line Interface"
    info = "For more information see: https://github.com/informatics-isi-edu/deriva-py"
    return DerivaTutorialCLI(desc, info).main()


if __name__ == '__main__':
    sys.exit(main())



import json
import logging
import requests
from datetime import datetime
from deriva.core import get_credential, get_new_requests_session, urlsplit, urlunsplit, format_exception, stob
from deriva.core.utils.webauthn_utils import get_wallet_entries
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import *


class GlobusWorkspacePortalPostProcessor(BaseProcessor):
    """
    Post processor that POSTs minids and metadata to the Globus Workspace Portal at:
    https://globus-portal.fair-research.org

    This processor is DEMO-ware and will be removed from the core transfer package after a few months.
    """

    WORKSPACE_API_URL = "https://globus-portal.fair-research.org/4M.4.Fullstacks/api/v1/workspaces/"

    def __init__(self, envars=None, **kwargs):
        super(GlobusWorkspacePortalPostProcessor, self).__init__(envars, **kwargs)

    def _get_access_token_from_wallet(self):
        entries = get_wallet_entries(
            self.wallet, "oauth2",
            credential_source="https://auth.globus.org",
            resource_server="fair_research_data_portal",
            scopes=["https://auth.globus.org/scopes/ebcaf30d-8148-4f1b-992a-bd089f823ac7/workspace_manager"])
        token = entries[0].get("access_token") if entries else None
        return token

    def process(self):
        url = self.parameters.get("url", self.WORKSPACE_API_URL)
        session = get_new_requests_session()
        headers = {"Content-Type": "application/json"}
        token = self._get_access_token_from_wallet()
        if token:
            headers.update({"Authorization": "Bearer %s" % token})
        tasks = self.parameters.get("tasks", ["WES", "JUPYTERHUB"])
        data_set = self.parameters.get("data_set")
        data_id = self.parameters.get("data_id")
        for k, v in self.outputs.items():
            minid = v[IDENTIFIER_KEY]
            logging.info("Registering minid [%s] with Workspace Portal: [%s]" % (minid, url))
            hostname = self.envars.get("hostname")
            hostname = "" if not hostname else " (%s)" % hostname
            metadata = self.parameters.get("metadata",
                                           {"grouping": "DERIVA%s" % hostname,
                                            "data_set": "%s" % data_set if data_set else k,
                                            "data_id": "%s" % data_id if data_id else datetime.now().isoformat()})
            workspaces = v.get("workspaces", [])
            body = {
                "input_minid": minid,
                "tasks": tasks,
                "metadata": metadata
            }
            try:
                response = session.post(url, json=body, headers=headers)
                response.raise_for_status()
                workspaces.append(response.json())
                v["workspaces"] = workspaces
            except Exception as e:
                raise DerivaDownloadError("Unable to create workspace: %s" % format_exception(e))

        return self.outputs



import os
from deriva.core import stob
from deriva.core.utils.mime_utils import guess_content_type
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import BaseProcessor, LOCAL_PATH_KEY, SOURCE_URL_KEY
from bdbag import bdbag_ro as ro


class BaseTransformProcessor(BaseProcessor):
    """
    Base class for TransformProcessor classes
    """

    def __init__(self, envars=None, **kwargs):
        super(BaseTransformProcessor, self).__init__(envars, **kwargs)
        self.base_path = kwargs["base_path"]
        self.input_path = self.parameters["input_path"]
        self.sub_path = self.parameters.get("output_path", "")
        self.is_bag = kwargs.get("bag", False)
        self.transformed_output = self.outputs.get(self.input_path, dict())
        self.url = self.transformed_output.get(SOURCE_URL_KEY)
        self.ro_file_provenance = stob(self.parameters.get("ro_file_provenance", False if not self.is_bag else True))
        self.ro_manifest = self.kwargs.get("ro_manifest")
        self.ro_author_name = self.kwargs.get("ro_author_name")
        self.ro_author_orcid = self.kwargs.get("ro_author_orcid")
        self.delete_input = stob(self.parameters.get("delete_input", True))
        self.input_relpath = None
        self.input_abspath = None
        self.output_relpath = None
        self.output_abspath = None

    def _create_input_output_paths(self):
        self.input_relpath, self.input_abspath = self.create_paths(
            self.base_path, self.input_path, is_bag=self.is_bag, envars=self.envars)
        self.output_relpath, self.output_abspath = self.create_paths(
            self.base_path, self.sub_path, is_bag=self.is_bag, envars=self.envars)

    def _delete_input(self):
        if os.path.isfile(self.input_abspath):
            os.remove(self.input_abspath)
        del self.outputs[self.input_relpath]

    def process(self):
        if self.ro_manifest and self.ro_file_provenance:
            ro.add_file_metadata(self.ro_manifest,
                                 source_url=self.url,
                                 local_path=self.output_relpath,
                                 media_type=guess_content_type(self.output_abspath),
                                 retrieved_on=ro.make_retrieved_on(),
                                 retrieved_by=ro.make_retrieved_by(self.ro_author_name, orcid=self.ro_author_orcid),
                                 bundled_as=ro.make_bundled_as())
        if self.delete_input:
            self._delete_input()

        self.outputs.update({self.output_relpath: {LOCAL_PATH_KEY: self.output_abspath, SOURCE_URL_KEY: self.url}})
        return self.outputs



